# Copyright 2023 Silicon Compiler Authors. All Rights Reserved.
import argparse
import sys
import os
import importlib
from inspect import getmembers, isfunction, getfullargspec
from siliconcompiler._metadata import version
from siliconcompiler.schema import utils


__default_source_file = "make.py"


def __process_file(path):
    if not os.path.exists(path):
        return {}, None, None
    mod_name = 'scmake'
    spec = importlib.util.spec_from_file_location(mod_name, path)
    make = importlib.util.module_from_spec(spec)
    sys.modules[mod_name] = make
    syspath = sys.path.copy()
    sys.path.insert(0, os.getcwd())
    spec.loader.exec_module(make)
    sys.path = syspath

    args = {}
    for name, func in getmembers(make, isfunction):
        if name.startswith('_'):
            continue

        # generate doc
        docstring = utils.trim(func.__doc__)
        if not docstring:
            docstring = f"run \"{name}\""
        short_help = docstring.splitlines()[0]

        func_spec = getfullargspec(func)

        func_args = {}
        for arg in func_spec.args:
            arg_type = str
            if arg in func_spec.annotations:
                arg_type = func_spec.annotations[arg]
            func_args[arg] = {
                "type": arg_type
            }

        if func_spec.defaults:
            for arg, defval in zip(reversed(func_spec.args), reversed(func_spec.defaults)):
                func_args[arg]["default"] = defval

                if defval is None:
                    continue

                if type(defval) is not func_args[arg]["type"]:
                    if isinstance(defval, (bool, str, float, int)):
                        func_args[arg]["type"] = type(defval)

        args[name] = {
            "function": func,
            "help": short_help,
            "full_help": docstring,
            "args": func_args
        }

    if args:
        default_arg = list(args.keys())[0]
    else:
        default_arg = None

    default_arg = getattr(make, '__scdefault', default_arg)

    module_help = utils.trim(make.__doc__)

    return args, default_arg, module_help


def main(source_file=None):
    progname = "smake"
    description = f"""-----------------------------------------------------------
SC app that provides an Makefile like interface to python
configuration files. This utility app will analyze a file
"{__default_source_file}" or the file specified with --file to determine
the available targets.

To view the help, use:
    smake --help

    or view the help for a specific target:
    smake --help <target>

To run a target, use:
    smake <target>

    or run a target from a file other than "{__default_source_file}":
    smake --file <file> <target>

    or run a target in a different directory:
    smake --directory <directory> <target>

To run a target with supported arguments, use:
    smake <target> --flow asicflow
-----------------------------------------------------------"""

    # handle source file identification before arg parse
    file_args = None
    if not source_file:
        source_file = __default_source_file
        file_args = ('--file', '-f')
        for file_arg in file_args:
            if file_arg in sys.argv:
                source_file_idx = sys.argv.index(file_arg) + 1
                if source_file_idx < len(sys.argv):
                    source_file = sys.argv[source_file_idx]
                else:
                    source_file = None
                break

    # handle directory identification before arg parse
    source_dir = os.getcwd()
    dir_args = ('--directory', '-C')
    for file_arg in dir_args:
        if file_arg in sys.argv:
            source_dir_idx = sys.argv.index(file_arg) + 1
            if source_dir_idx < len(sys.argv):
                source_dir = sys.argv[source_dir_idx]
            else:
                source_dir = None
            break

    if source_dir:
        if not os.path.isdir(source_dir):
            print(f"Unable to change directory to {source_dir}")
            return 1

        os.chdir(source_dir)

    make_args = {}
    default_arg = None
    module_help = None
    if source_file:
        make_args, default_arg, module_help = __process_file(source_file)

    if module_help:
        description += \
            f"\n\n{module_help}\n\n-----------------------------------------------------------"

    parser = argparse.ArgumentParser(
        progname,
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    if file_args:
        parser.add_argument(
            *file_args,
            metavar='<file>',
            help=f'Use file as makefile, default is {__default_source_file}')

    parser.add_argument(
        *dir_args,
        metavar='<directory>',
        help='Change to directory <directory> before reading the makefile.')

    parser.add_argument(
        '--version', '-v',
        action='version',
        version=version)

    targetparsers = parser.add_subparsers(
        dest='target',
        metavar='<target>',
        help='Target to execute')

    for arg, info in make_args.items():
        subparse = targetparsers.add_parser(
            arg,
            description=info['full_help'],
            help=info['help'],
            formatter_class=argparse.RawDescriptionHelpFormatter)

        for subarg, subarg_info in info['args'].items():
            # print(subarg, subarg_info)
            add_args = {}

            if "default" not in subarg_info:
                add_args["required"] = True
            else:
                if type(subarg_info["default"]) is subarg_info["type"]:
                    add_args["default"] = subarg_info["default"]

            if subarg_info["type"] is bool:
                def str2bool(v):
                    # modified from:
                    # https://github.com/pypa/distutils/blob/8993718731b951ee36d08cb784f02aa13542ce15/distutils/util.py
                    val = v.lower()
                    if val in ('y', 'yes', 't', 'true', 'on', '1'):
                        return True
                    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
                        return False
                    else:
                        raise ValueError(f"invalid truth value {val!r}")
                subarg_info["type"] = str2bool

            subparse.add_argument(
                f'--{subarg}',
                dest=f'sub_{subarg}',
                metavar=f'<{subarg}>',
                type=subarg_info["type"],
                **add_args)

    args = parser.parse_args()
    target = args.target
    if not target:
        target = default_arg

    if not os.path.isfile(source_file):
        print(f"Unable to load {source_file}")
        return 1

    call_args = {}
    args_vars = vars(args)
    for arg in make_args[target]["args"]:
        if f'sub_{arg}' in args_vars and args_vars[f'sub_{arg}'] is not None:
            call_args[arg] = args_vars[f'sub_{arg}']
    make_args[target]["function"](**call_args)

    return 0


#########################
if __name__ == "__main__":
    sys.exit(main())



# Copyright 2023 Silicon Compiler Authors. All Rights Reserved.

from siliconcompiler import Chip


class PackageChip(Chip):
    def __init__(self, *args, package=None):
        # Start with None as init setting will not depend on package
        self.__package = None

        name = args[-1]
        super().__init__(name)

        if len(args) == 2:
            self.logger.warning(
                f'passing Chip object to {name} ({type(self).__name__}) is deprecated')

        path = None
        ref = None
        if isinstance(package, (tuple, list)):
            if len(package) == 3:
                package, path, ref = package
            elif len(package) == 2:
                package, path = package
            else:
                raise ValueError(f"{package} should be a 2 or 3 item tuple or list.")
        elif isinstance(package, dict):
            if len(package) == 1:
                info = list(package.values())[0]
                if "path" not in info:
                    raise ValueError(f"{package} should contain a path key.")
                path = info["path"]
                if "ref" in info:
                    ref = info["ref"]

                package = list(package.keys())[0]
            else:
                raise ValueError(f"{package} cannot contain multiple packages.")
        elif isinstance(package, str):
            pass
        else:
            if package is not None:
                raise ValueError(f"{package} is not supported.")

        if path:
            self.register_source(package, path, ref=ref)

        self.__package = package

        # Clear all copy flags since these are libraries, pdks, fpga, etc.
        for key in self.allkeys():
            sc_type = self.get(*key, field='type')
            if 'file' in sc_type or 'dir' in sc_type:
                self.set(*key, False, field='copy')

    def add(self, *args, field='value', step=None, index=None, package=None):
        if not package:
            package = self.__package
        super().add(*args, field=field, step=step,
                    index=index, package=package)

    def set(self, *args, field='value', clobber=True, step=None, index=None, package=None):
        if not package:
            package = self.__package
        super().set(*args, field=field, clobber=clobber, step=step,
                    index=index, package=package)

    def input(self, filename, fileset=None, filetype=None, iomap=None,
              step=None, index=None, package=None):
        self._add_input_output('input', filename, fileset, filetype, iomap,
                               step=step, index=index, package=package,
                               quiet=True)

    def output(self, filename, fileset=None, filetype=None, iomap=None,
               step=None, index=None, package=None):
        self._add_input_output('output', filename, fileset, filetype, iomap,
                               step=step, index=index, package=package,
                               quiet=True)


class PDK(PackageChip):
    """
    Object for configuring a process development kit.
    This is the main object used for configuration and data for a PDK
    within the SiliconCompiler platform.

    This inherits all methods from :class:`~siliconcompiler.Chip`.

    Args:
        name (string): Name of the PDK.
        package (string): Name of the data source
    Examples:
        >>> siliconcompiler.PDK("asap7")
        Creates a flow object with name "asap7".
    """


class FPGA(PackageChip):
    """
    Object for configuring an FPGA
    This is the main object used for configuration and data for a FPGA
    within the SiliconCompiler platform.

    This inherits all methods from :class:`~siliconcompiler.Chip`.

    Args:
        name (string): Name of the FPGA.
        package (string): Name of the data source
    Examples:
        >>> siliconcompiler.FPGA("lattice_ice40")
        Creates a flow object with name "lattice_ice40".
    """


class Library(PackageChip):
    """
    Object for configuring a library.
    This is the main object used for configuration and data for a library
    within the SiliconCompiler platform.

    This inherits all methods from :class:`~siliconcompiler.Chip`.

    Args:
        name (string): Name of the library.
        package (string): Name of the data source
        auto_enable (boolean): If True, will automatically be added to ['option','library'].
            This is only valid for non-logiclibs and macrolibs
    Examples:
        >>> siliconcompiler.Library("asap7sc7p5t")
        Creates a library object with name "asap7sc7p5t".
    """
    def __init__(self, *args, package=None, auto_enable=False):
        super().__init__(*args, package=package)

        self.__auto_enable = auto_enable

    def is_auto_enable(self):
        return self.__auto_enable


class Flow(Chip):
    """
    Object for configuring a flow.
    This is the main object used for configuration and data for a flow
    within the SiliconCompiler platform.

    This inherits all methods from :class:`~siliconcompiler.Chip`.

    Args:
        name (string): Name of the flow.
    Examples:
        >>> siliconcompiler.Flow("asicflow")
        Creates a flow object with name "asicflow".
    """
    def __init__(self, *args):
        super().__init__(args[-1])
        if len(args) == 2:
            self.logger.warning(f'passing Chip object to {type(self)} is deprecated')


class Checklist(Chip):
    """
    Object for configuring a checklist.
    This is the main object used for configuration and data for a checklist
    within the SiliconCompiler platform.

    This inherits all methods from :class:`~siliconcompiler.Chip`.

    Args:
        name (string): Name of the checklist.
    Examples:
        >>> siliconcompiler.Checklist("tapeout")
        Creates a checklist object with name "tapeout".
    """
    def __init__(self, *args):
        super().__init__(args[-1])
        if len(args) == 2:
            self.logger.warning(f'passing Chip object to {type(self)} is deprecated')



from siliconcompiler import utils
from siliconcompiler.tools import slang
from siliconcompiler.tools._common import \
    add_require_input, add_frontend_requires, get_tool_task


def setup(chip):
    '''
    Lint system verilog
    '''
    slang.setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             clobber=False, step=step, index=index)

    add_require_input(chip, 'input', 'rtl', 'verilog')
    add_require_input(chip, 'input', 'rtl', 'systemverilog')
    add_frontend_requires(chip, ['ydir', 'idir', 'vlib', 'libext', 'define', 'param'])


def runtime_options(chip):
    options = slang.common_runtime_options(chip)
    options.extend([
        "--lint-only"
    ])

    return options


def post_process(chip):
    slang.post_process(chip)



import siliconcompiler

from siliconcompiler.flows._common import setup_multiple_frontends
from siliconcompiler.flows._common import _make_docs

from siliconcompiler.tools.yosys import syn_asic
from siliconcompiler.tools.opensta import timing

from siliconcompiler.tools.builtin import minimum


############################################################################
# DOCS
############################################################################
def make_docs(chip):
    n = 3
    _make_docs(chip)
    return setup(syn_np=n, timing_np=n)


###########################################################################
# Flowgraph Setup
############################################################################
def setup(flowname='synflow',
          syn_np=1,
          timing_np=1):
    '''
    A configurable ASIC synthesys flow with static timing.

    The 'synflow' includes the stages below. The steps syn have
    minimization associated with them.
    To view the flowgraph, see the .png file.

    * **import**: Sources are collected and packaged for compilation
    * **syn**: Translates RTL to netlist using Yosys
    * **timing**: Create timing reports of design

    The syn and timing steps supports per process
    options that can be set up by setting 'syn_np' or 'timing_np'
    arg to a value > 1, as detailed below:

    * syn_np : Number of parallel synthesis jobs to launch
    * timing_np : Number of parallel timing jobs to launch
    '''

    flow = siliconcompiler.Flow(flowname)

    # Linear flow, up until branch to run parallel verification steps.
    longpipe = ['syn',
                'synmin',
                'timing']

    # step --> task
    tasks = {
        'syn': syn_asic,
        'synmin': minimum,
        'timing': timing
    }

    np = {
        "syn": syn_np,
        "timing": timing_np
    }

    prevstep = None
    # Remove built in steps where appropriate
    flowpipe = []
    for step in longpipe:
        task = tasks[step]
        if task == minimum:
            if prevstep in np and np[prevstep] > 1:
                flowpipe.append(step)
        else:
            flowpipe.append(step)
        prevstep = step

    flowtasks = []
    for step in flowpipe:
        flowtasks.append((step, tasks[step]))

    # Programmatically build linear portion of flowgraph and fanin/fanout args
    prevstep = setup_multiple_frontends(flow)
    for step, task in flowtasks:
        fanout = 1
        if step in np:
            fanout = np[step]
        # create nodes
        for index in range(fanout):
            # nodes
            flow.node(flowname, step, task, index=index)

            # edges
            if task == minimum:
                fanin = 1
                if prevstep in np:
                    fanin = np[prevstep]
                for i in range(fanin):
                    flow.edge(flowname, prevstep, step, tail_index=i)
            elif prevstep:
                flow.edge(flowname, prevstep, step, head_index=index)

            # metrics
            goal_metrics = ()
            weight_metrics = ()
            if task in (syn_asic, ):
                goal_metrics = ('errors',)
                weight_metrics = ()
            elif task in (timing, ):
                goal_metrics = ('errors', 'setupwns', 'setuptns')
                weight_metrics = ('cellarea', 'peakpower', 'leakagepower')

            for metric in goal_metrics:
                flow.set('flowgraph', flowname, step, str(index), 'goal', metric, 0)
            for metric in weight_metrics:
                flow.set('flowgraph', flowname, step, str(index), 'weight', metric, 1.0)
        prevstep = step

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    chip.set('input', 'constraint', 'sdc', 'test')
    flow = make_docs(chip)
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



import re

from siliconcompiler.tools.magic.magic import setup as setup_tool
from siliconcompiler.tools.magic.magic import process_file
from siliconcompiler import sc_open
from siliconcompiler.tools._common import get_tool_task, record_metric
from siliconcompiler.tools._common.asic import get_mainlib, get_libraries


def setup(chip):
    '''
    Perform DRC checks
    '''

    # Generic tool setup
    setup_tool(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    chip.set('tool', tool, 'task', task, 'output', f'{design}.drc.mag', step=step, index=index)


def pre_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    pdk = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')
    mainlib = get_mainlib(chip)
    libtype = chip.get('library', mainlib, 'asic', 'libarch', step=step, index=index)
    process_file('lef', chip, 'pdk', pdk, 'aprtech', 'magic', stackup, libtype, 'lef')

    for lib in get_libraries(chip, 'logic'):
        process_file('lef', chip, 'library', lib, 'output', stackup, 'lef')

    for lib in get_libraries(chip, 'macro'):
        if lib in chip.get('tool', tool, 'task', task, 'var', 'exclude', step=step, index=index):
            process_file('lef', chip, 'library', lib, 'output', stackup, 'lef')


################################
# Post_process (post executable)
################################
def post_process(chip):
    ''' Tool specific function to run after step execution

    Reads error count from output and fills in appropriate entry in metrics
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    design = chip.top()

    report_path = f'reports/{design}.drc'
    drcs = 0
    with sc_open(report_path) as f:
        for line in f:
            errors = re.search(r'^\[INFO\]: COUNT: (\d+)', line)

            if errors:
                drcs = errors.group(1)
    record_metric(chip, step, index, 'drcs', drcs, report_path)

    # TODO: return error code
    return 0




'''
ImageMagick® is a free and open-source software suite for displaying, converting, and editing raster
image and vector image files.
It can read and write over 200 image file formats, and can support a wide range of image
manipulation operations, such as resizing, cropping, and color correction.
Use the montage program to create a composite image by combining several separate images.
The images are tiled on the composite image optionally adorned with a border, frame, image name,
and more

Documentation: https://imagemagick.org/

Sources: https://github.com/ImageMagick/ImageMagick

Installation: https://github.com/ImageMagick/ImageMagick
'''


def setup(chip):
    exe = 'montage'

    chip.set('tool', 'montage', 'exe', exe)
    chip.set('tool', 'montage', 'vswitch', '-version')
    chip.set('tool', 'montage', 'version', '>=6.9.0')


def parse_version(stdout):
    first_line = stdout.splitlines()[0]

    return first_line.split(' ')[2]



from siliconcompiler import NodeStatus

from siliconcompiler.tools._common import has_pre_post_script, get_tool_task

from siliconcompiler.tools.openroad._apr import setup as apr_setup
from siliconcompiler.tools.openroad._apr import set_reports, set_pnr_inputs, set_pnr_outputs
from siliconcompiler.tools.openroad._apr import \
    define_ord_params, define_sta_params, define_sdc_params, \
    define_mpl_params, define_gpl_params
from siliconcompiler.tools.openroad._apr import build_pex_corners, define_ord_files
from siliconcompiler.tools.openroad._apr import extract_metrics


def setup(chip):
    '''
    Macro placement
    '''

    # Generic apr tool setup.
    apr_setup(chip)

    # Task setup
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'script', 'apr/sc_macro_placement.tcl',
             step=step, index=index)

    # Setup task IO
    set_pnr_inputs(chip)
    set_pnr_outputs(chip)

    # set default values for openroad
    define_ord_params(chip)
    define_sta_params(chip)
    define_sdc_params(chip)
    define_mpl_params(chip)
    define_gpl_params(chip)

    set_reports(chip, [
        'setup',
        'unconstrained'
    ])

    chip.set('tool', tool, 'task', task, 'file', 'rtlmp_constraints',
             'contraints script for macro placement',
             field='help')

    if chip.get('tool', tool, 'task', task, 'file', 'rtlmp_constraints', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'file', 'rtlmp_constraints']),
                 step=step, index=index)


def pre_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    input_nodes = chip.get('record', 'inputnode', step=step, index=index)
    if not has_pre_post_script(chip) and all([
            chip.get('metric', 'macros', step=in_step, index=in_index) == 0
            for in_step, in_index in input_nodes
            ]):
        chip.set('record', 'status', NodeStatus.SKIPPED, step=step, index=index)
        chip.logger.warning(f'{step}{index} will be skipped since are no macros to place.')
        return

    build_pex_corners(chip)
    define_ord_files(chip)


def post_process(chip):
    extract_metrics(chip)



import os
import re
import shutil
from siliconcompiler.utils import sc_open
from siliconcompiler.tools._common.asic import set_tool_task_var, set_tool_task_lib_var, get_mainlib
from siliconcompiler.tools._common.asic_clock import get_clock_period, add_clock_requirements
from siliconcompiler.tools._common import \
    add_frontend_requires, add_require_input, get_frontend_options, get_input_files, \
    get_tool_task, has_input_files, record_metric


def make_docs(chip):
    from siliconcompiler.targets import freepdk45_demo
    chip.use(freepdk45_demo)
    chip.input('<design>.c')
    return setup(chip)


def setup(chip):
    '''
    Performs high level synthesis to generate a verilog output
    '''

    if not has_input_files(chip, 'input', 'hll', 'c') and \
       not has_input_files(chip, 'input', 'hll', 'llvm'):
        return "no files in [input,hll,c] or [input,hll,llvm]"

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    # Standard Setup
    refdir = 'tools/' + tool
    chip.set('tool', tool, 'exe', 'bambu')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=2024.03', clobber=False)

    chip.set('tool', tool, 'task', task, 'refdir', refdir,
             step=step, index=index,
             package='siliconcompiler', clobber=False)

    # Input/Output requirements
    chip.add('tool', tool, 'task', task, 'output', chip.top() + '.v', step=step, index=index)

    add_clock_requirements(chip)

    # Schema requirements
    add_require_input(chip, 'input', 'hll', 'c')
    add_require_input(chip, 'input', 'hll', 'llvm')
    add_frontend_requires(chip, ['idir', 'define'])

    set_tool_task_var(chip, 'device',
                      schelp="Device to use during bambu synthesis")
    set_tool_task_lib_var(chip, 'memorychannels', default_value=1,
                          schelp="Number of memory channels available")

    # Require clock conversion factor, from library units to ns
    mainlib = get_mainlib(chip)
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['library', mainlib, 'option', 'var', 'bambu_clock_multiplier']),
             step=step, index=index)

    set_tool_task_var(chip, 'clock_multiplier',
                      schelp="Clock multiplier used to convert library units to ns")


################################
#  Custom runtime options
################################
def runtime_options(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    cmdlist = []

    opts = get_frontend_options(chip, ['idir', 'define'])

    for value in opts['idir']:
        cmdlist.append('-I' + value)
    for value in opts['define']:
        cmdlist.append('-D' + value)
    for value in get_input_files(chip, 'input', 'hll', 'c'):
        cmdlist.append(value)
    if not has_input_files(chip, 'input', 'hll', 'c'):
        # Only use llvm if C is empty
        for value in get_input_files(chip, 'input', 'hll', 'llvm'):
            cmdlist.append(value)

    cmdlist.append('--soft-float')
    cmdlist.append('--memory-allocation-policy=NO_BRAM')

    mem_channels = int(chip.get('tool', tool, 'task', task, 'var', 'memorychannels',
                                step=step, index=index)[0])
    if mem_channels > 0:
        cmdlist.append(f'--channels-number={mem_channels}')

    mainlib = get_mainlib(chip)
    clock_multiplier = float(chip.get('library', mainlib, 'option', 'var',
                                      'bambu_clock_multiplier')[0])
    clock_name, period = get_clock_period(chip, clock_units_multiplier=clock_multiplier)
    if clock_name:
        cmdlist.append(f'--clock-name={clock_name}')
    if period:
        cmdlist.append(f'--clock-period={period}')

    cmdlist.append('--disable-function-proxy')

    device = chip.get('tool', tool, 'task', task, 'var', 'device',
                      step=step, index=index)
    if device:
        cmdlist.append(f'--device={device[0]}')

    cmdlist.append(f'--top-fname={chip.top(step, index)}')

    return cmdlist


################################
# Post_process (post executable)
################################
def post_process(chip):
    ''' Tool specific function to run after step execution
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    shutil.copy2(f'{chip.top(step, index)}.v', os.path.join('outputs', f'{chip.top()}.v'))

    ff = re.compile(fr"Total number of flip-flops in function {chip.top(step, index)}: (\d+)")
    area = re.compile(r"Total estimated area: (\d+)")
    fmax = re.compile(r"Estimated max frequency \(MHz\): (\d+\.?\d*)")
    slack = re.compile(r"Minimum slack: (\d+\.?\d*)")

    log_file = f"{step}.log"
    with sc_open(log_file) as log:
        for line in log:
            ff_match = ff.findall(line)
            area_match = area.findall(line)
            fmax_match = fmax.findall(line)
            slack_match = slack.findall(line)
            if ff_match:
                record_metric(chip, step, index, "registers", int(ff_match[0]), log_file)
            if area_match:
                record_metric(chip, step, index, "cellarea", float(area_match[0]), log_file,
                              source_unit='um^2')
            if fmax_match:
                record_metric(chip, step, index, "fmax", float(fmax_match[0]), log_file,
                              source_unit='MHz')
            if slack_match:
                slack_ns = float(slack_match[0])
                if slack_ns >= 0:
                    record_metric(chip, step, index, "setupwns", 0, log_file,
                                  source_unit='ns')
                else:
                    record_metric(chip, step, index, "setupwns", slack_ns, log_file,
                                  source_unit='ns')
                record_metric(chip, step, index, "setupslack", slack_ns, log_file,
                              source_unit='ns')



from siliconcompiler._common import NodeStatus, SiliconCompilerError

from siliconcompiler.utils import sc_open

from siliconcompiler.core import Chip
from siliconcompiler.schema import Schema

from siliconcompiler._metadata import version as __version__

from siliconcompiler.use import PDK, FPGA, Library, Flow, Checklist

__all__ = [
    "__version__",
    "Chip",
    "SiliconCompilerError",
    "NodeStatus",
    "PDK",
    "FPGA",
    "Library",
    "Flow",
    "Checklist",
    "Schema",
    "sc_open"
]



import siliconcompiler

from siliconcompiler.flows._common import setup_multiple_frontends
from siliconcompiler.flows._common import _make_docs

from siliconcompiler.tools.yosys import syn_asic
from siliconcompiler.tools.openroad import init_floorplan
from siliconcompiler.tools.openroad import macro_placement
from siliconcompiler.tools.openroad import endcap_tapcell_insertion
from siliconcompiler.tools.openroad import power_grid
from siliconcompiler.tools.openroad import pin_placement
from siliconcompiler.tools.openroad import global_placement
from siliconcompiler.tools.openroad import repair_design
from siliconcompiler.tools.openroad import detailed_placement
from siliconcompiler.tools.openroad import clock_tree_synthesis
from siliconcompiler.tools.openroad import repair_timing
from siliconcompiler.tools.openroad import fillercell_insertion
from siliconcompiler.tools.openroad import global_route
from siliconcompiler.tools.openroad import antenna_repair
from siliconcompiler.tools.openroad import detailed_route
from siliconcompiler.tools.openroad import fillmetal_insertion
from siliconcompiler.tools.openroad import write_data
from siliconcompiler.tools.klayout import export as klayout_export

from siliconcompiler.tools.builtin import minimum


############################################################################
# DOCS
############################################################################
def make_docs(chip):
    n = 3
    _make_docs(chip)
    return setup(syn_np=n, floorplan_np=n, physyn_np=n, place_np=n, cts_np=n, route_np=n)


###########################################################################
# Flowgraph Setup
############################################################################
def setup(flowname='asicflow',
          syn_np=1,
          floorplan_np=1,
          physyn_np=1,
          place_np=1,
          cts_np=1,
          route_np=1):
    '''
    A configurable ASIC compilation flow.

    The 'asicflow' includes the stages below. The steps syn, floorplan,
    physyn, place, cts, route, and dfm have minimization associated
    with them. To view the flowgraph, see the .png file.

    * **import**: Sources are collected and packaged for compilation
    * **syn**: Translates RTL to netlist using Yosys
    * **floorplan**: Floorplanning
    * **physyn**: Physical Synthesis
    * **place**: Global and detailed placement
    * **cts**: Clock tree synthesis
    * **route**: Global and detailed routing
    * **dfm**: Metal fill, atenna fixes and any other post routing steps
    * **export**: Export design from APR tool and merge with library GDS
    * **sta**: Static timing analysis (signoff)
    * **lvs**: Layout versus schematic check (signoff)
    * **drc**: Design rule check (signoff)

    The syn, physyn, place, cts, route steps supports per process
    options that can be set up by setting '<step>_np'
    arg to a value > 1, as detailed below:

    * syn_np : Number of parallel synthesis jobs to launch
    * floorplan_np : Number of parallel floorplan jobs to launch
    * physyn_np : Number of parallel physical synthesis jobs to launch
    * place_np : Number of parallel place jobs to launch
    * cts_np : Number of parallel clock tree synthesis jobs to launch
    * route_np : Number of parallel routing jobs to launch
    '''

    flow = siliconcompiler.Flow(flowname)

    # Linear flow, up until branch to run parallel verification steps.
    longpipe = [
        'syn',
        'syn.min',
        'floorplan.init',
        'floorplan.macro_placement',
        'floorplan.tapcell',
        'floorplan.power_grid',
        'floorplan.pin_placement',
        'floorplan.min',
        'place.global',
        'place.repair_design',
        'place.detailed',
        'place.min',
        'cts.clock_tree_synthesis',
        'cts.repair_timing',
        'cts.fillcell',
        'cts.min',
        'route.global',
        'route.antenna_repair',
        'route.detailed',
        'route.min',
        'dfm.metal_fill'
    ]

    # step --> task
    tasks = {
        'syn': syn_asic,
        'syn.min': minimum,
        'floorplan.init': init_floorplan,
        'floorplan.macro_placement': macro_placement,
        'floorplan.tapcell': endcap_tapcell_insertion,
        'floorplan.power_grid': power_grid,
        'floorplan.pin_placement': pin_placement,
        'floorplan.min': minimum,
        'place.global': global_placement,
        'place.repair_design': repair_design,
        'place.detailed': detailed_placement,
        'place.min': minimum,
        'cts.clock_tree_synthesis': clock_tree_synthesis,
        'cts.repair_timing': repair_timing,
        'cts.fillcell': fillercell_insertion,
        'cts.min': minimum,
        'route.global': global_route,
        'route.antenna_repair': antenna_repair,
        'route.detailed': detailed_route,
        'route.min': minimum,
        'dfm.metal_fill': fillmetal_insertion
    }

    np = {
        "syn": syn_np,
        "floorplan": floorplan_np,
        "physyn": physyn_np,
        "place": place_np,
        "cts": cts_np,
        "route": route_np
    }

    prevstep = None
    # Remove built in steps where appropriate
    flowpipe = []
    for step in longpipe:
        task = tasks[step]
        if task == minimum:
            np_step = prevstep.split('.')[0]
            if np_step in np and np[np_step] > 1:
                flowpipe.append(step)
        else:
            flowpipe.append(step)
        prevstep = step

    flowtasks = []
    for step in flowpipe:
        flowtasks.append((step, tasks[step]))

    # Programmatically build linear portion of flowgraph and fanin/fanout args
    prevstep = setup_multiple_frontends(flow)
    prev_fanout = 1
    for step, task in flowtasks:
        fanout = 1
        np_step = step.split('.')[0]
        if np_step in np and task != minimum:
            fanout = np[np_step]

        # create nodes
        for index in range(fanout):
            # nodes
            flow.node(flowname, step, task, index=index)

        # create edges
        for index in range(fanout):
            # edges
            fanin = prev_fanout
            if task == minimum:
                for i in range(fanin):
                    flow.edge(flowname, prevstep, step, tail_index=i)
            elif prevstep:
                if fanin == fanout:
                    flow.edge(flowname, prevstep, step, tail_index=index, head_index=index)
                else:
                    flow.edge(flowname, prevstep, step, head_index=index)

            # metrics
            goal_metrics = ()
            weight_metrics = ()
            if task in (syn_asic, ):
                goal_metrics = ('errors',)
                weight_metrics = ()
            elif task in (init_floorplan, macro_placement, endcap_tapcell_insertion,
                          power_grid, pin_placement, global_placement, repair_design,
                          detailed_placement, clock_tree_synthesis, repair_timing,
                          fillercell_insertion, global_route, antenna_repair, detailed_route,
                          fillmetal_insertion):
                goal_metrics = ('errors', 'setupwns', 'setuptns')
                weight_metrics = ('cellarea', 'peakpower', 'leakagepower')

            for metric in goal_metrics:
                flow.set('flowgraph', flowname, step, str(index), 'goal', metric, 0)
            for metric in weight_metrics:
                flow.set('flowgraph', flowname, step, str(index), 'weight', metric, 1.0)
        prevstep = step
        prev_fanout = fanout

    # add write information steps
    flow.node(flowname, 'write.gds', klayout_export)
    flow.edge(flowname, prevstep, 'write.gds')
    flow.node(flowname, 'write.views', write_data)
    flow.edge(flowname, prevstep, 'write.views')

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    chip.set('input', 'constraint', 'sdc', 'test')
    flow = make_docs(chip)
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



from siliconcompiler.tools.klayout.klayout import setup as setup_tool
from siliconcompiler.tools.klayout.show import general_gui_setup
from siliconcompiler.tools.klayout.show import pre_process as show_pre_process
from siliconcompiler.tools.klayout.show import runtime_options as show_runtime_options
from siliconcompiler.tools._common import get_tool_task


def make_docs(chip):
    from siliconcompiler.tools.klayout import klayout
    klayout.make_docs(chip)
    chip.set('tool', 'klayout', 'task', 'screenshot', 'var', 'show_filepath', '<path>')


def setup(chip):
    '''
    Generate a PNG file from a layout file
    '''

    # Generic tool setup.
    setup_tool(chip)

    tool = 'klayout'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    clobber = False

    chip.set('tool', tool, 'task', task, 'threads', 1, step=step, index=index, clobber=clobber)

    setup_gui_screenshot(chip)

    option = ['-nc', '-z', '-rm']
    chip.set('tool', tool, 'task', task, 'option', option, step=step, index=index, clobber=clobber)


def pre_process(chip):
    show_pre_process(chip)


def setup_gui_screenshot(chip, require_input=True):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    design = chip.top()

    general_gui_setup(chip, task, True, require_input=require_input)

    chip.set('tool', tool, 'task', task, 'var', 'show_horizontal_resolution', '4096',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'show_vertical_resolution', '4096',
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'xbins', '1',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'ybins', '1',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'margin', '10',
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'linewidth', '0',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'oversampling', '2',
             step=step, index=index, clobber=False)

    # Help
    chip.set('tool', tool, 'task', task, 'var', 'show_horizontal_resolution',
             'Horizontal resolution in pixels',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'show_vertical_resolution',
             'Vertical resolution in pixels',
             field='help')

    chip.set('tool', tool, 'task', task, 'var', 'xbins',
             'If greater than 1, splits the image into multiple segments along x-axis',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'ybins',
             'If greater than 1, splits the image into multiple segments along y-axis',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'margin',
             'Margin around design in microns',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'linewidth',
             'Width of lines in detailed screenshots',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'oversampling',
             'Image oversampling used in detailed screenshots',
             field='help')

    xbins = int(chip.get('tool', tool, 'task', task, 'var', 'xbins',
                         step=step, index=index)[0])
    ybins = int(chip.get('tool', tool, 'task', task, 'var', 'ybins',
                         step=step, index=index)[0])

    if xbins == 1 and ybins == 1:
        chip.add('tool', tool, 'task', task, 'output', design + '.png',
                 step=step, index=index)
    else:
        for x in range(xbins):
            for y in range(ybins):
                chip.add('tool', tool, 'task', task, 'output', f'{design}_X{x}_Y{y}.png',
                         step=step, index=index)


def runtime_options(chip):
    return show_runtime_options(chip)



from siliconcompiler.flows import asicflow
from siliconcompiler.flows import asictopflow
from siliconcompiler.flows import drcflow
from siliconcompiler.flows import dvflow
from siliconcompiler.flows import fpgaflow
from siliconcompiler.flows import generate_openroad_rcx
from siliconcompiler.flows import interposerflow
from siliconcompiler.flows import lintflow
from siliconcompiler.flows import screenshotflow
from siliconcompiler.flows import showflow
from siliconcompiler.flows import signoffflow
from siliconcompiler.flows import synflow


def get_flows():
    '''
    Returns a dict of builtin flows
    '''
    return {
        module.__name__.split(".")[-1]: module for module in (
            asicflow,
            asictopflow,
            drcflow,
            dvflow,
            fpgaflow,
            generate_openroad_rcx,
            interposerflow,
            lintflow,
            screenshotflow,
            showflow,
            signoffflow,
            synflow
        )
    }



# Copyright 2020 Silicon Compiler Authors. All Rights Reserved.

from aiohttp import web
import threading
import json
import logging as log
import os
import shutil
import uuid
import tarfile
import sys
import fastjsonschema
from pathlib import Path
from fastjsonschema import JsonSchemaException
import io

from siliconcompiler import Chip, Schema
from siliconcompiler.schema import utils as schema_utils
from siliconcompiler._metadata import version as sc_version
from siliconcompiler.schema import SCHEMA_VERSION as sc_schema_version
from siliconcompiler.remote.schema import ServerSchema
from siliconcompiler.remote import banner, JobStatus
from siliconcompiler.scheduler.slurm import get_configuration_directory
from siliconcompiler.flowgraph import nodes_to_execute


# Compile validation code for API request bodies.
api_dir = Path(__file__).parent / 'server_schema' / 'requests'

# 'remote_run': Run a stage of a job using the server's cluster settings.
with open(api_dir / 'remote_run.json') as schema:
    validate_remote_run = fastjsonschema.compile(json.loads(schema.read()))

# 'check_progress': Check whether a given job stage is currently running.
with open(api_dir / 'check_progress.json') as schema:
    validate_check_progress = fastjsonschema.compile(json.loads(schema.read()))

# 'check_server': Check whether a given job stage is currently running.
with open(api_dir / 'check_server.json') as schema:
    validate_check_server = fastjsonschema.compile(json.loads(schema.read()))

# 'cancel_job': Cancel a running job.
with open(api_dir / 'cancel_job.json') as schema:
    validate_cancel_job = fastjsonschema.compile(json.loads(schema.read()))

# 'delete_job': Delete a job and remove it from server-side storage.
with open(api_dir / 'delete_job.json') as schema:
    validate_delete_job = fastjsonschema.compile(json.loads(schema.read()))

# 'get_results': Fetch the results of a job run.
# Currently, the 'job_hash' is included in the URL for this call.
with open(api_dir / 'get_results.json') as schema:
    validate_get_results = fastjsonschema.compile(json.loads(schema.read()))


class Server:
    """
    The core class for the siliconcompiler 'gateway' server, which can run
    locally or on a remote host. Its job is to process requests for
    asynchronous siliconcompiler jobs, by using the slurm HPC daemon to
    schedule work on available compute nodes. It can also be configured to
    launch new compute nodes in the cloud, for on-demand jobs.
    """

    __version__ = '0.0.1'

    ####################
    def __init__(self, loglevel="info"):
        '''
        Init method for Server object
        '''

        # Initialize logger
        self.logger = log.getLogger(f'sc_server_{id(self)}')
        handler = log.StreamHandler(stream=sys.stdout)
        formatter = log.Formatter('%(asctime)s | %(levelname)-8s | %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(schema_utils.translate_loglevel(loglevel))

        self.schema = ServerSchema(logger=self.logger)

        # Set up a dictionary to track running jobs.
        self.sc_jobs = {}

    def run(self):
        if not os.path.exists(self.nfs_mount):
            raise FileNotFoundError(f'{self.nfs_mount} could not be found.')

        self.logger.info(f"Running in: {self.nfs_mount}")
        # If authentication is enabled, try connecting to the SQLite3 database.
        # (An empty one will be created if it does not exist.)

        # If authentication is enabled, load [username : password/key] mappings.
        # Remember, this development server is intended to be a minimal
        # demonstration of the API, and should only be used for testing purposes.
        # You should NEVER store plaintext passwords in a production
        # server implementation.
        if self.get('option', 'auth'):
            self.user_keys = {}
            json_path = os.path.join(self.nfs_mount, 'users.json')
            try:
                with open(json_path, 'r') as users_file:
                    users_json = json.loads(users_file.read())
                for mapping in users_json['users']:
                    username = mapping['username']
                    self.user_keys[username] = {
                        'password': mapping['password'],
                        'compute_time': 0,
                        'bandwidth': 0
                    }
                    if 'compute_time' in mapping:
                        self.user_keys[username]['compute_time'] = mapping['compute_time']
                    if 'bandwidth' in mapping:
                        self.user_keys[username]['bandwidth'] = mapping['bandwidth']
            except Exception:
                self.logger.warning("Could not find well-formatted 'users.json' "
                                    "file in the server's working directory. "
                                    "(User : Key) mappings were not imported.")

        # Create a minimal web server to process the 'remote_run' API call.
        self.app = web.Application()
        self.app.add_routes([
            web.post('/remote_run/', self.handle_remote_run),
            web.post('/check_progress/', self.handle_check_progress),
            web.post('/check_server/', self.handle_check_server),
            web.post('/delete_job/', self.handle_delete_job),
            web.post('/get_results/{job_hash}.tar.gz', self.handle_get_results),
        ])
        # TODO: Put zip files in a different directory.
        # For security reasons, this is not a good public-facing solution.
        # There's no access control on which files can be downloaded.
        # But this is an example server which only implements a minimal API.
        self.app.router.add_static('/get_results/', self.nfs_mount)

        # Start the async server.
        web.run_app(self.app, port=self.get('option', 'port'))

    def create_cmdline(self, progname, description=None, switchlist=None, additional_args=None):
        def print_banner():
            print(banner)
            print("Version:", Server.__version__, "\n")
            print("-" * 80)

        return self.schema.create_cmdline(
            progname=progname,
            description=description,
            switchlist=switchlist,
            input_map=None,
            additional_args=additional_args,
            version=Server.__version__,
            print_banner=print_banner,
            logger=self.logger)

    ####################
    async def handle_remote_run(self, request):
        '''
        API handler for 'remote_run' commands. This method delegates
        a 'Chip.run(...)' method to a compute node using slurm.
        '''

        # Temporary file path to store streamed data.
        tmp_file = os.path.join(self.nfs_mount, uuid.uuid4().hex)

        # Set up a multipart reader to read in the large file, and param data.
        reader = await request.multipart()
        while True:
            # Get the next part; if it doesn't exist, we're done.
            part = await reader.next()
            if part is None:
                break

            # Save the initial 'import' step archive. Note: production server
            # implementations may want to encrypt data before storing it on disk.
            if part.name == 'import':
                with open(tmp_file, 'wb') as f:
                    while True:
                        chunk = await part.read_chunk()
                        if not chunk:
                            break
                        f.write(chunk)

            # Retrieve JSON request parameters.
            elif part.name == 'params':
                # Get the job parameters.
                params = await part.json()

                if 'chip_cfg' not in params:
                    return self.__response('Manifest not provided.', status=400)
                chip_cfg = params['chip_cfg']

        # Process input parameters
        job_params, response = self._check_request(params['params'],
                                                   validate_remote_run)
        if response is not None:
            return response

        # Create a dummy Chip object to make schema traversal easier.
        # start with a dummy name, as this will be overwritten
        chip = Chip('server')
        # Add provided schema
        chip.schema = Schema(cfg=chip_cfg)

        # Fetch some common values.
        design = chip.design
        job_name = chip.get('option', 'jobname')
        job_hash = uuid.uuid4().hex
        chip.set('record', 'remoteid', job_hash)

        # Ensure that the job's root directory exists.
        job_root = os.path.join(self.nfs_mount, job_hash)
        job_dir = os.path.join(job_root, design, job_name)
        os.makedirs(job_dir, exist_ok=True)

        # Move the uploaded archive and un-zip it.
        # (Contents will be encrypted for authenticated jobs)
        with tarfile.open(tmp_file, "r:gz") as tar:
            tar.extractall(path=job_dir)

        # Delete the temporary file if it still exists.
        if os.path.exists(tmp_file):
            os.remove(tmp_file)

        # Create the working directory for the given 'job hash' if necessary.
        chip.set('option', 'builddir', job_root)

        # Remove 'remote' JSON config value to run locally on compute node.
        chip.set('option', 'remote', False)

        # Write JSON config to shared compute storage.
        os.makedirs(os.path.join(job_root, 'configs'), exist_ok=True)

        # Run the job with the configured clustering option. (Non-blocking)
        job_proc = threading.Thread(target=self.remote_sc,
                                    args=[
                                        chip,
                                        job_params['username']])
        job_proc.start()

        # Return a response to the client.
        return web.json_response({'message': f"Starting job: {job_hash}",
                                  'interval': self.checkinterval,
                                  'job_hash': job_hash})

    ####################
    async def handle_get_results(self, request):
        '''
        API handler to redirect 'get_results' POST calls.
        '''

        # Process input parameters
        params = await request.json()
        job_params, response = self._check_request(params,
                                                   validate_get_results)
        job_params['job_hash'] = request.match_info.get('job_hash', '')
        if response is not None:
            return response

        job_hash = job_params['job_hash']
        node = job_params['node'] if 'node' in job_params else None

        resp = web.StreamResponse(
            status=200,
            reason='OK',
            headers={
                'Content-Type': 'application/x-tar',
                'Content-Disposition': f'attachment; filename="{job_hash}_{node}.tar.gz"'
            },
        )
        await resp.prepare(request)

        zipfn = os.path.join(self.nfs_mount, job_hash, f'{job_hash}_{node}.tar.gz')
        if not node:
            with tarfile.open(zipfn, 'w:gz') as tar:
                text = "Done"
                metadata_file = io.BytesIO(text.encode('ascii'))
                tarinfo = tarfile.TarInfo(f'{job_hash}/done')
                tarinfo.size = metadata_file.getbuffer().nbytes
                tar.addfile(tarinfo=tarinfo, fileobj=metadata_file)

        with open(zipfn, 'rb') as zipf:
            await resp.write(zipf.read())

        await resp.write_eof()

        return resp

    ####################
    async def handle_delete_job(self, request):
        '''
        API handler for 'delete_job' requests. Delete a job from shared
        cloud compute storage.
        '''

        # Process input parameters
        job_params, response = self._check_request(await request.json(),
                                                   validate_delete_job)
        if response is not None:
            return response

        job_hash = job_params['job_hash']

        # Determine if the job is running.
        for job in self.sc_jobs:
            if job_hash in job:
                return self.__response("Error: job is still running.", status=400)

        # Delete job hash directory, only if it exists.
        # TODO: This assumes no malicious input.
        # Again, we will need a more mature server framework to implement
        # good access control and security policies for a public-facing service.
        build_dir = os.path.join(self.nfs_mount, job_hash)
        check_dir = os.path.dirname(build_dir)
        if check_dir == self.nfs_mount:
            if os.path.exists(build_dir):
                shutil.rmtree(build_dir)

            tar_file = f'{build_dir}.tar.gz'
            if os.path.exists(tar_file):
                os.remove(tar_file)

        return web.Response(text="Job deleted.")

    ####################
    async def handle_check_progress(self, request):
        '''
        API handler for the 'check progress' endpoint. Currently,
        It only returns a response containing a 'still running', 'done', or
        'not found' message. In the future, it can respond with up-to-date
        information about the job's progress and intermediary outputs.
        '''

        # Process input parameters
        job_params, response = self._check_request(await request.json(),
                                                   validate_check_progress)
        if response is not None:
            return response

        job_hash = job_params['job_hash']
        username = job_params['username']

        jobname = self.job_name(username, job_hash)

        # Determine if the job is running.
        # TODO: Return information about individual flowgraph nodes.
        if jobname in self.sc_jobs:
            resp = {
                'status': JobStatus.RUNNING,
                'message': 'Job is currently running on the server.',
            }
        else:
            resp = {
                'status': JobStatus.COMPLETED,
                'message': 'Job has no running steps.',
            }
        return web.json_response(resp)

    ####################
    async def handle_check_server(self, request):
        '''
        API handler for the 'check user' endpoint.
        '''

        # Process input parameters

        job_params, response = self._check_request(await request.json(),
                                                   validate_check_server)
        if response is not None:
            return response

        resp = {
            'status': 'ready',
            'versions': {
                'sc': sc_version,
                'sc_schema': sc_schema_version,
                'sc_server': Server.__version__,
            },
            'progress_interval': self.checkinterval
        }

        username = job_params['username']
        if username:
            resp['user_info'] = {
                'compute_time': self.user_keys[username]['compute_time'],
                'bandwidth_kb': self.user_keys[username]['bandwidth'],
            }

        return web.json_response(resp)

    def job_name(self, username, job_hash):
        if username:
            return f'{username}_{job_hash}'
        else:
            return job_hash

    ####################
    def remote_sc(self, chip, username):
        '''
        Async method to delegate an '.run()' command to a host,
        and send an email notification when the job completes.
        '''

        # Assemble core job parameters.
        job_hash = chip.get('record', 'remoteid')

        # Mark the job run as busy.
        sc_job_name = self.job_name(username, job_hash)
        self.sc_jobs[sc_job_name] = 'busy'

        build_dir = os.path.join(self.nfs_mount, job_hash)
        chip.set('option', 'builddir', build_dir)
        chip.set('option', 'remote', False)

        job_cfg_dir = get_configuration_directory(chip)
        os.makedirs(job_cfg_dir, exist_ok=True)
        chip.write_manifest(f"{job_cfg_dir}/chip{chip.get('option', 'jobname')}.json")

        if self.get('option', 'cluster') == 'slurm':
            # Run the job with slurm clustering.
            chip.set('option', 'scheduler', 'name', 'slurm')

        # Run the job.
        chip.run()

        # Archive each task.
        for (step, index) in nodes_to_execute(chip):
            chip.cwd = os.path.join(chip.get('option', 'builddir'), '..')
            tf = tarfile.open(os.path.join(self.nfs_mount,
                                           job_hash,
                                           f'{job_hash}_{step}{index}.tar.gz'),
                              mode='w:gz')
            chip._archive_node(tf, step=step, index=index)
            tf.close()

        # (Email notifications can be sent here using your preferred API)

        # Mark the job hash as being done.
        self.sc_jobs.pop(sc_job_name)

    ####################
    def __auth_password(self, username, password):
        '''
        Helper method to authenticate a username : password combination.
        This minimal implementation of the API uses a simple string match, because
        we don't have a database/signup process/etc for the development server.
        But production server implementations should ALWAYS hash and salt
        passwords before storing them.
        NEVER store production passwords as plaintext!
        '''

        if username not in self.user_keys:
            return False
        return password == self.user_keys[username]['password']

    def _check_request(self, request, json_validator):
        params = {}
        if request:
            try:
                params = json_validator(request)
            except JsonSchemaException as e:
                return (params, self.__response(f"Error: Invalid parameters: {e}.", status=400))

            if not request:
                return (params, self.__response("Error: Invalid parameters.", status=400))

        # Check for authentication parameters.
        if self.get('option', 'auth'):
            if ('username' in params) and ('key' in params):
                # Authenticate the user.
                if not self.__auth_password(params['username'], params['key']):
                    return (params, self.__response("Authentication error.", status=403))

            else:
                return (params,
                        self.__response("Error: some authentication parameters are missing.",
                                        status=400))

        if 'username' not in params:
            params['username'] = None

        return (params, None)

    ###################
    def __response(self, message, status=200):
        return web.json_response({'message': message}, status=status)

    ###################
    @property
    def nfs_mount(self):
        # Ensure that NFS mounting path is absolute.
        return os.path.abspath(self.get('option', 'nfsmount'))

    ###################
    @property
    def checkinterval(self):
        return self.get('option', 'checkinterval')

    def get(self, *keypath, field='value'):
        return self.schema.get(*keypath, field=field)

    def set(self, *args, field='value', clobber=True):
        keypath = args[:-1]
        value = args[-1]

        if keypath == ['option', 'loglevel'] and field == 'value':
            self.logger.setLevel(schema_utils.translate_loglevel(value))

        self.schema.set(*keypath, value, field=field, clobber=clobber)

    def write_configuration(self, filepath):
        with open(filepath, 'w') as f:
            self.schema.write_json(f)



from siliconcompiler._common import NodeStatus as SCNodeStatus

# CLI entry banner autogenerated using pyfiglet.
# >> pyfiglet.figlet_format("SC Server")
banner = r'''
 ____   ____   ____
/ ___| / ___| / ___|  ___ _ ____   _____ _ __
\___ \| |     \___ \ / _ \ '__\ \ / / _ \ '__|
 ___) | |___   ___) |  __/ |   \ V /  __/ |
|____/ \____| |____/ \___|_|    \_/ \___|_|
'''


class NodeStatus(SCNodeStatus):
    '''
    Enum class to help ensure consistent status messages
    '''

    # special code for uploaded
    UPLOADED = 'uploaded'


class JobStatus():
    '''
    Enum class to help ensure consistent status messages
    '''

    RUNNING = "running"

    COMPLETED = "completed"
    FAILED = "failed"
    CANCELED = "canceled"
    REJECTED = "rejected"
    TIMEOUT = "timeout"

    UNKNOWN = "unknown"


from siliconcompiler.remote.client import Client, ConfigureClient # noqa E402


__all__ = [
    "NodeStatus",
    "JobStatus",
    "Client",
    "ConfigureClient"
]



from siliconcompiler import NodeStatus

from siliconcompiler.tools._common import get_tool_task, has_pre_post_script
from siliconcompiler.tools._common.asic import set_tool_task_var

from siliconcompiler.tools.openroad._apr import setup as apr_setup
from siliconcompiler.tools.openroad._apr import set_reports, set_pnr_inputs, set_pnr_outputs
from siliconcompiler.tools.openroad._apr import \
    define_ord_params, define_sta_params, define_sdc_params, \
    define_pdn_params, define_psm_params
from siliconcompiler.tools.openroad._apr import build_pex_corners, \
    define_ord_files, define_pdn_files
from siliconcompiler.tools.openroad._apr import extract_metrics


def setup(chip):
    '''
    Perform power grid insertion and connectivity analysis
    '''

    # Generic apr tool setup.
    apr_setup(chip)

    # Task setup
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'script', 'apr/sc_power_grid.tcl',
             step=step, index=index)

    # Setup task IO
    set_pnr_inputs(chip)
    set_pnr_outputs(chip)

    # set default values for openroad
    define_ord_params(chip)
    define_sta_params(chip)
    define_sdc_params(chip)
    define_pdn_params(chip)
    define_psm_params(chip)

    set_tool_task_var(chip, param_key='fixed_pin_keepout',
                      default_value=0,
                      schelp='if > 0, applies a blockage in multiples of the routing pitch '
                             'to each fixed pin to ensure there is room for routing.')

    set_reports(chip, [])


def pre_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    define_pdn_files(chip)
    pdncfg = [file for file in chip.find_files('tool', tool, 'task', task, 'file', 'pdn_config',
                                               step=step, index=index) if file]
    if not has_pre_post_script(chip) and \
            (chip.get('tool', tool, 'task', task, 'var', 'pdn_enable',
                      step=step, index=index)[0] == 'false' or len(pdncfg) == 0):
        chip.set('record', 'status', NodeStatus.SKIPPED, step=step, index=index)
        chip.logger.warning(f'{step}{index} will be skipped since power grid is disabled.')
        return

    define_ord_files(chip)
    build_pex_corners(chip)


def post_process(chip):
    extract_metrics(chip)



import shutil
from siliconcompiler import utils
from siliconcompiler import SiliconCompilerError
from siliconcompiler.tools.vpr import vpr
from siliconcompiler.tools.vpr._json_constraint import load_constraints_map
from siliconcompiler.tools.vpr._json_constraint import load_json_constraints
from siliconcompiler.tools.vpr._json_constraint import map_constraints
from siliconcompiler.tools.vpr._xml_constraint import generate_vpr_constraints_xml_file
from siliconcompiler.tools._common import get_tool_task


def setup(chip, clobber=True):
    '''
    Perform automated place and route with VPR
    '''

    tool = 'vpr'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    vpr.setup_tool(chip, clobber=clobber)

    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index, clobber=False)

    design = chip.top()
    chip.set('tool', tool, 'task', task, 'input', design + '.blif', step=step, index=index)

    chip.set('tool', tool, 'task', task, 'output', design + '.blif', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.net', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.place', step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'placement', 'component placement constraints',
             field='help')


def runtime_options(chip):
    '''Command line options to vpr for the place step
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    options = vpr.runtime_options(chip)

    blif = f"inputs/{design}.blif"
    options.append(blif)

    options.append('--pack')
    options.append('--place')

    enable_images = chip.get('tool', tool, 'task', task, 'var', 'enable_images',
                             step=step, index=index)[0]

    if enable_images == 'true':
        graphics_commands = vpr.get_common_graphics(chip)

        graphics_command_str = " ".join(graphics_commands)

        options.append("--graphics_commands")
        options.append(f"\"{graphics_command_str}\"")

    return options


################################
# Pre_process (pre executable)
################################


def pre_process(chip):
    ''' Tool specific function to run before step execution
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    part_name = chip.get('fpga', 'partname')

    # If the user explicitly provides an XML constraints file, give that
    # priority over other constraints input types:
    if chip.valid('input', 'constraint', 'vpr_pins'):
        constraint_file = vpr.find_single_file(chip, 'input', 'constraint', 'vpr_pins',
                                               step=step, index=index,
                                               file_not_found_msg="VPR constraints file not found")

        if (constraint_file is not None):
            shutil.copy2(constraint_file, vpr.auto_constraints())

    elif chip.valid('input', 'constraint', 'pcf'):
        constraint_file = vpr.find_single_file(chip, 'input', 'constraint', 'pcf',
                                               step=step, index=index,
                                               file_not_found_msg="PCF constraints file not found")

        map_file = vpr.find_single_file(chip, 'fpga', part_name, 'file', 'constraints_map',
                                        file_not_found_msg="constraints map not found")

        if not map_file:
            raise SiliconCompilerError('FPGA does not have required constraints map', chip=chip)

        constraints_map = load_constraints_map(map_file)
        json_constraints = load_json_constraints(constraint_file)
        all_place_constraints, missing_pins = map_constraints(chip,
                                                              json_constraints,
                                                              constraints_map)
        if (missing_pins > 0):
            raise SiliconCompilerError(
                "Pin constraints specify I/O ports not in this architecture", chip=chip)

        generate_vpr_constraints_xml_file(all_place_constraints, vpr.auto_constraints())

    else:
        all_component_constraints = chip.get('tool', tool, 'task', task, 'var', 'placement',
                                             step=step, index=index)

        all_place_constraints = {}
        for constraint in all_component_constraints:
            component, *place_constraint = constraint.split(",")
            place_constraint = tuple([int(loc) for loc in place_constraint])
            chip.logger.info(f'Place constraint for {component} at {place_constraint}')
            all_place_constraints[component] = place_constraint

        if all_place_constraints:
            generate_vpr_constraints_xml_file(all_place_constraints, vpr.auto_constraints())


def add_placement_constraint(chip, component, location, step=None, index=None):
    tool = 'vpr'
    task = 'place'

    constraint = f"{component},{','.join([str(loc) for loc in location])}"

    chip.add('tool', tool, 'task', task, 'var', 'placement', constraint, step=step, index=index)

################################
# Post_process (post executable)
################################


def post_process(chip):
    ''' Tool specific function to run after step execution
    '''
    vpr.vpr_post_process(chip)

    design = chip.top()
    shutil.copy2(f'inputs/{design}.blif', 'outputs')



import os
import shlex
import subprocess
import stat
import time
import uuid
import json
import shutil
from siliconcompiler import utils, SiliconCompilerError
from siliconcompiler.flowgraph import nodes_to_execute, _get_flowgraph_entry_nodes
from siliconcompiler.package import get_cache_path

# Full list of Slurm states, split into 'active' and 'inactive' categories.
# Many of these do not apply to a minimal configuration, but we'll track them all.
# https://slurm.schedmd.com/squeue.html#SECTION_JOB-STATE-CODES
SLURM_ACTIVE_STATES = [
    'RUNNING',
    'PENDING',
    'CONFIGURING',
    'COMPLETING',
    'SIGNALING',
    'STAGE_OUT',
    'RESIZING',
    'REQUEUED',
]
SLURM_INACTIVE_STATES = [
    'BOOT_FAIL',
    'CANCELLED',
    'COMPLETED',
    'DEADLINE',
    'FAILED',
    'NODE_FAIL',
    'OUT_OF_MEMORY',
    'PREEMPTED',
    'RESV_DEL_HOLD',
    'REQUEUE_FED',
    'REQUEUE_HOLD',
    'REVOKED',
    'SPECIAL_EXIT',
    'STOPPED',
    'SUSPENDED',
    'TIMEOUT',
]


###########################################################################
def get_configuration_directory(chip):
    '''
    Helper function to get the configuration directory for the scheduler
    '''

    return f'{chip.getworkdir()}/configs'


def init(chip):
    if os.path.exists(chip._getcollectdir()):
        # nothing to do
        return

    collect = False
    flow = chip.get('option', 'flow')
    entry_nodes = _get_flowgraph_entry_nodes(chip, flow)
    for (step, index) in nodes_to_execute(chip, flow):
        if (step, index) in entry_nodes:
            collect = True

    if collect:
        chip.collect()


###########################################################################
def _defernode(chip, step, index, replay):
    '''
    Helper method to run an individual step on a slurm cluster.

    Blocks until the compute node
    finishes processing this step, and it sets the active/error bits.
    '''

    # Determine which HPC job scheduler being used.
    scheduler_type = chip.get('option', 'scheduler', 'name', step=step, index=index)

    if scheduler_type != 'slurm':
        raise ValueError(f'{scheduler_type} is not a supported scheduler')

    if not check_slurm():
        raise SiliconCompilerError('slurm is not available or installed on this machine', chip=chip)

    # Determine which cluster parititon to use. (Default value can be overridden on per-step basis)
    partition = chip.get('option', 'scheduler', 'queue', step=step, index=index)
    if not partition:
        partition = _get_slurm_partition()

    # Get the temporary UID associated with this job run.
    job_hash = chip.get('record', 'remoteid')
    if not job_hash:
        # Generate a new uuid since it was not set
        job_hash = uuid.uuid4().hex

    job_name = f'{job_hash}_{step}{index}'

    # Write out the current schema for the compute node to pick up.
    cfg_dir = get_configuration_directory(chip)
    cfg_file = f'{cfg_dir}/{step}{index}.json'
    log_file = f'{cfg_dir}/{step}{index}.log'
    script_file = f'{cfg_dir}/{step}{index}.sh'
    os.makedirs(cfg_dir, exist_ok=True)

    chip.set('option', 'scheduler', 'name', None, step=step, index=index)
    chip.write_manifest(cfg_file)

    # Allow user-defined compute node execution script if it already exists on the filesystem.
    # Otherwise, create a minimal script to run the task using the SiliconCompiler CLI.
    if not os.path.isfile(script_file):
        with open(script_file, 'w') as sf:
            sf.write(utils.get_file_template('slurm/run.sh').render(
                cfg_file=shlex.quote(cfg_file),
                build_dir=shlex.quote(chip.get("option", "builddir")),
                step=shlex.quote(step),
                index=shlex.quote(index),
                cachedir=shlex.quote(get_cache_path(chip))
            ))

    # This is Python for: `chmod +x [script_path]`
    os.chmod(script_file,
             os.stat(script_file).st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    schedule_cmd = ['sbatch',
                    '--exclusive',
                    '--partition', partition,
                    '--chdir', chip.cwd,
                    '--job-name', job_name,
                    '--output', log_file]

    # Only delay the starting time if the 'defer' Schema option is specified.
    defer_time = chip.get('option', 'scheduler', 'defer', step=step, index=index)
    if defer_time:
        schedule_cmd.extend(['--begin', defer_time])

    schedule_cmd.append(script_file)

    # Run the 'srun' command, and track its output.
    # TODO: output should be fed to log, and stdout if quiet = False
    step_result = subprocess.Popen(schedule_cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT)

    # Wait for the subprocess call to complete. It should already be done,
    # as it has closed its output stream. But if we don't call '.wait()',
    # the '.returncode' value will not be set correctly.
    step_result.wait()
    result_msg = step_result.stdout.read().decode()
    sbatch_id = result_msg.split(' ')[-1].strip()
    retcode = 0

    while True:
        # Return early with an error if the batch ID is not an integer.
        if not sbatch_id.isdigit():
            retcode = 1
            break

        # Rate-limit the status checks to once every few seconds.
        time.sleep(3.0)

        # Check whether the job is still running.
        jobcheck = subprocess.run(['scontrol', 'show', 'job', sbatch_id],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)
        jobout = jobcheck.stdout.decode()

        # Jobs have a number of potential states that they can be in if they
        # are still active in the Slurm scheduler.
        if [state for state in SLURM_ACTIVE_STATES if state in jobout]:
            pass
        # 'COMPLETED' is a special case indicating successful job termination.
        elif 'COMPLETED' in jobout:
            break
        elif 'Invalid job id specified' in jobout:
            # May have already completed and been purged from active list.
            break
        # Jobs have a number of potential states that they can be in if they
        # did not terminate successfully.
        elif [state for state in SLURM_INACTIVE_STATES if state in jobout]:
            # FAILED, TIMEOUT, etc.
            retcode = 1
            break

    if retcode > 0:
        chip.logger.error(f'srun command for {step} failed.')
        chip.logger.error(f'srun output for {step}: {jobout}')


def _get_slurm_partition():
    partitions = subprocess.run(['sinfo', '--json'],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

    if partitions.returncode != 0:
        raise RuntimeError('Unable to determine partitions in slurm')

    sinfo = json.loads(partitions.stdout.decode())

    # Return the first listed partition
    return sinfo['nodes'][0]['partitions'][0]


def check_slurm():
    return shutil.which('sinfo') is not None



'''
OpenSTA is a gate level static timing verifier.

Documentation: https://github.com/The-OpenROAD-Project/OpenSTA/blob/master/doc/OpenSTA.pdf

Sources: https://github.com/The-OpenROAD-Project/OpenSTA

Installation: https://github.com/The-OpenROAD-Project/OpenSTA (also installed with OpenROAD)
'''

from siliconcompiler import utils
from siliconcompiler.tools.openroad._apr import get_library_timing_keypaths
from siliconcompiler.tools._common import get_tool_task
from siliconcompiler.tools._common.asic import get_libraries


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    from siliconcompiler.targets import asap7_demo
    chip.use(asap7_demo)


################################
# Setup Tool (pre executable)
################################
def setup(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'exe', 'sta')
    chip.set('tool', tool, 'vswitch', '-version')
    chip.set('tool', tool, 'version', '>=v2.5.0', clobber=False)
    chip.set('tool', tool, 'format', 'tcl')

    targetlibs = get_libraries(chip, 'logic')
    macrolibs = get_libraries(chip, 'macro')
    delaymodel = chip.get('asic', 'delaymodel', step=step, index=index)

    # Input/Output requirements for default asicflow steps
    chip.set('tool', tool, 'task', task, 'refdir', 'tools/opensta/scripts',
             step=step, index=index,
             package='siliconcompiler', clobber=False)
    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index, clobber=False)

    if delaymodel != 'nldm':
        chip.logger.error(f'{delaymodel} delay model is not supported by {tool}, only nldm')

    if targetlibs:
        # Note: only one footprint supported in mainlib
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['asic', 'logiclib']),
                 step=step, index=index)

        for lib in targetlibs:
            for timing_key in get_library_timing_keypaths(chip, lib).values():
                chip.add('tool', tool, 'task', task, 'require', ",".join(timing_key),
                         step=step, index=index)
        for lib in macrolibs:
            for timing_key in get_library_timing_keypaths(chip, lib).values():
                if chip.valid(*timing_key):
                    chip.add('tool', tool, 'task', task, 'require', ",".join(timing_key),
                             step=step, index=index)
    else:
        chip.error('logiclib parameters required for OpenSTA.')

    # basic warning and error grep check on logfile
    chip.set('tool', tool, 'task', task, 'regex', 'warnings', r'^\[WARNING|^Warning',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'errors', r'^\[ERROR',
             step=step, index=index, clobber=False)


################################
# Version Check
################################
def parse_version(stdout):
    return stdout.strip()


################################
# Runtime options
################################
def runtime_options(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    # exit automatically in batch mode and not breakpoint
    option = []
    if not chip.get('option', 'breakpoint', step=step, index=index):
        option.append("-exit")

    tool, task = get_tool_task(chip, step, index)
    option.extend([
        '-threads',
        str(chip.get('tool', tool, 'task', task, 'threads', step=step, index=index))
    ])

    return option



from siliconcompiler.targets import asap7_demo
from siliconcompiler.targets import asic_demo
from siliconcompiler.targets import fpgaflow_demo
from siliconcompiler.targets import freepdk45_demo
from siliconcompiler.targets import gf180_demo
from siliconcompiler.targets import ihp130_demo
from siliconcompiler.targets import interposer_demo
from siliconcompiler.targets import skywater130_demo


def get_targets():
    '''
    Returns a dict of builtin targets
    '''
    return {
        module.__name__.split(".")[-1]: module for module in (
            asap7_demo,
            asic_demo,
            fpgaflow_demo,
            freepdk45_demo,
            gf180_demo,
            ihp130_demo,
            interposer_demo,
            skywater130_demo
        )
    }



'''
Bluespec is a high-level hardware description language. It has a variety of
advanced features including a powerful type system that can prevent errors
prior to synthesis time, and its most distinguishing feature, Guarded Atomic
Actions, allow you to define hardware components in a modular manner based
on their invariants, and let the compiler pick a scheduler.

Documentation: https://github.com/B-Lang-org/bsc#documentation

Sources: https://github.com/B-Lang-org/bsc

Installation: https://github.com/B-Lang-org/bsc#download
'''

from siliconcompiler.tools.bluespec import convert


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    convert.setup(chip)
    return chip


################################
# Setup Tool (pre executable)
################################
def parse_version(stdout):
    # Examples:
    # Bluespec Compiler, version 2021.12.1-27-g9a7d5e05 (build 9a7d5e05)
    # Bluespec Compiler, version 2021.07 (build 4cac6eba)

    long_version = stdout.split()[3]
    return long_version.split('-')[0]



import graphviz
import os

from siliconcompiler import sc_open
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Generate a screenshot of a dot file
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'threads', 1, step=step, index=index)

    chip.add('tool', tool, 'task', task, 'output', chip.top() + '.png', step=step, index=index)


def run(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    if os.path.exists(f'inputs/{chip.top()}.dot'):
        file = f'inputs/{chip.top()}.dot'
    elif os.path.exists(f'inputs/{chip.top()}.xdot'):
        file = f'inputs/{chip.top()}.xdot'
    elif chip.valid('tool', tool, 'task', task, 'var', 'show_filepath') and \
            chip.get('tool', tool, 'task', task, 'var', 'show_filepath', step=step, index=index):
        file = chip.get('tool', tool, 'task', task, 'var', 'show_filepath',
                        step=step, index=index)[0]
    else:
        file = chip.find_files('input', 'image', 'dot', step=step, index=index)[0]

    with sc_open(file) as dot:
        dot_content = dot.read()

    try:
        dot = graphviz.Source(dot_content, format="png")
        dot.render(filename=f"outputs/{chip.top()}", cleanup=True)
        pass
    except graphviz.ExecutableNotFound:
        return 1

    return 0



from . import get_libraries as common_get_libraries
from . import get_tool_task, pick_key
import json


def get_mainlib(chip):
    return get_libraries(chip, 'logic')[0]


def get_libraries(chip, type):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    if type not in ('logic', 'macro'):
        raise ValueError(f'Cannot collect {type} libraries')

    libs = []
    for lib in chip.get('asic', f'{type}lib', step=step, index=index):
        if lib in libs:
            continue
        libs.append(lib)

    for lib in common_get_libraries(chip, include_asic=False):
        if not chip.valid('library', lib, 'asic', f'{type}lib'):
            continue
        for sublib in chip.get('library', lib, 'asic', f'{type}lib', step=step, index=index):
            if sublib in libs:
                continue
            libs.append(sublib)

    return libs


def get_timing_modes(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    modes = set()
    for scenario in chip.getkeys('constraint', 'timing'):
        modes.add(chip.get('constraint', 'timing', scenario, 'mode',
                  step=step, index=index))
    return sorted(modes)


def set_tool_task_var(chip,
                      param_key,
                      default_value=None,
                      schelp=None,
                      option_key=None,
                      pdk_key=None,
                      lib_key=None,
                      require=None,
                      skip=None):
    '''
    Set parameter from PDK -> main library -> option -> default_value
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    pdkname = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    if not skip:
        skip = []
    if not isinstance(skip, (list, tuple)):
        skip = [skip]

    if not require:
        require = []
    if not isinstance(require, (list, tuple)):
        require = [require]

    check_keys = []

    # Add PDK key
    if 'pdk' not in skip:
        if not pdk_key:
            pdk_key = param_key
        check_keys.append(['pdk', pdkname, 'var', tool, stackup, pdk_key])
        if 'pdk' in require:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(check_keys[-1]),
                     step=step, index=index)

    # Add library key
    if 'lib' not in skip:
        mainlib = get_mainlib(chip)
        if not lib_key:
            lib_key = f'{tool}_{param_key}'
        check_keys.append(['library', mainlib, 'option', 'var', lib_key])
        if 'lib' in require:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(check_keys[-1]),
                     step=step, index=index)

    # Add option key
    if 'option' not in skip:
        if not option_key:
            option_key = f'{tool}_{param_key}'
        check_keys.append(['option', 'var', option_key])
        if 'option' in require:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(check_keys[-1]),
                     step=step, index=index)

    def check_value(val):
        if isinstance(val, (list, tuple)):
            return len(val) > 0
        return val is not None

    require_key, value = pick_key(chip, reversed(check_keys), step=step, index=index)
    if not check_value(value):
        value = default_value

    has_value = check_value(value)

    if has_value:
        chip.set('tool', tool, 'task', task, 'var', param_key, value,
                 step=step, index=index, clobber=False)

        if require_key:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(require_key),
                     step=step, index=index)

    if has_value or 'key' in require:
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', param_key]),
                 step=step, index=index)

    if schelp:
        chip.set('tool', tool, 'task', task, 'var', param_key,
                 schelp, field='help')

    return value


def set_tool_task_lib_var(chip,
                          param_key,
                          default_value=None,
                          schelp=None,
                          option_key=None,
                          lib_key=None):
    '''
    Set parameter from libraries -> option -> default_value
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    if schelp:
        chip.set('tool', tool, 'task', task, 'var', param_key,
                 schelp, field='help')

    if not option_key:
        option_key = f'{tool}_{param_key}'
    require_key, value = pick_key(chip, [('option', 'var', option_key)], step=step, index=index)

    def check_value(val):
        if isinstance(val, (list, tuple, set)):
            return len(val) > 0
        return val is not None

    if check_value(value):
        chip.set('tool', tool, 'task', task, 'var', param_key, value,
                 step=step, index=index, clobber=False)

        if require_key:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(('option', option_key)),
                     step=step, index=index)

        return value

    # Add library key
    if not lib_key:
        lib_key = f'{tool}_{param_key}'
    lib_keys = []
    for lib in get_libraries(chip, 'logic'):
        if chip.valid('library', lib, 'option', 'var', lib_key) and \
           chip.get('library', lib, 'option', 'var', lib_key):
            lib_keys.append(('library', lib, 'option', 'var', lib_key))

    values = set()
    for lib_key in lib_keys:
        chip.add('tool', tool, 'task', task, 'require', ','.join(lib_key), step=step, index=index)

        get_step = step
        get_index = index

        if chip.get(*lib_key, field='pernode').is_never():
            get_step = None
            get_index = None

        values.update(chip.get(*lib_key, step=get_step, index=get_index))

    if default_value and not check_value(values):
        values = default_value

    if check_value(values):
        chip.set('tool', tool, 'task', task, 'var', param_key, values,
                 step=step, index=index, clobber=False)

        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', param_key]),
                 step=step, index=index)

    return values


def get_tool_task_var(chip,
                      param_key,
                      option_key=None,
                      pdk_key=None,
                      lib_key=None,
                      skip=None):
    '''
    Get parameter from PDK -> main library -> option -> default_value
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, _ = get_tool_task(chip, step, index)
    pdkname = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    if not skip:
        skip = []
    if not isinstance(skip, (list, tuple)):
        skip = [skip]

    check_keys = []
    # Add PDK key
    if 'pdk' not in skip:
        if not pdk_key:
            pdk_key = param_key
        check_keys.append(['pdk', pdkname, 'var', tool, stackup, pdk_key])

    # Add library key
    if 'lib' not in skip:
        mainlib = get_mainlib(chip)
        if not lib_key:
            lib_key = f'{tool}_{param_key}'
        check_keys.append(['library', mainlib, 'option', 'var', lib_key])

    # Add option key
    if 'option' not in skip:
        if not option_key:
            option_key = f'{tool}_{param_key}'
        check_keys.append(['option', 'var', option_key])

    _, value = pick_key(chip, reversed(check_keys), step=step, index=index)

    return value


class CellArea:
    def __init__(self):
        self.__areas = {}

    def addCell(self, name=None, module=None,
                cellarea=None, cellcount=None,
                macroarea=None, macrocount=None,
                stdcellarea=None, stdcellcount=None):
        if not name and not module:
            return

        if all([metric is None for metric in (
                cellarea, cellcount,
                macroarea, macrocount,
                stdcellarea, stdcellcount)]):
            return

        if not name:
            name = module

        # ensure name is unique
        check_name = name
        idx = 0
        while check_name in self.__areas:
            check_name = f'{name}{idx}'
            idx += 1
        name = check_name

        self.__areas[name] = {
            "module": module,
            "cellarea": cellarea,
            "cellcount": cellcount,
            "macroarea": macroarea,
            "macrocount": macrocount,
            "stdcellarea": stdcellarea,
            "stdcellcount": stdcellcount
        }

    def size(self):
        return len(self.__areas)

    def writeReport(self, path):
        with open(path, 'w') as f:
            json.dump(self.__areas, f, indent=4)



'''Sphinx extension that provides directives for automatically generating
documentation for dynamically loaded modules used by SC.
'''

from docutils import nodes
from sphinx.util.nodes import nested_parse_with_titles
from docutils.statemachine import ViewList
from sphinx.util.docutils import SphinxDirective
from sphinx.domains.std import StandardDomain
from sphinx.addnodes import pending_xref
import docutils
from siliconcompiler.utils import get_plugins

import importlib
import pkgutil
import os
import subprocess

import siliconcompiler
from siliconcompiler.sphinx_ext import sc_root as SC_ROOT
from siliconcompiler.schema import Schema, utils
from siliconcompiler.sphinx_ext.utils import (
    strong,
    code,
    para,
    keypath,
    build_table,
    build_list,
    build_section,
    build_section_with_target,
    link,
    image,
    get_ref_id,
    literalblock
)

#############
# Helpers
#############


def build_schema_value_table(cfg, refdoc, keypath_prefix=None, skip_zero_weight=False):
    '''Helper function for displaying values set in schema as a docutils table.'''
    table = [[strong('Keypath'), strong('Value')]]

    # Nest received dictionary under keypath_prefix
    rooted_cfg = cfg
    if keypath_prefix:
        for key in reversed(keypath_prefix):
            rooted_cfg = {key: rooted_cfg}

    def format_value(is_list, value):
        if is_list:
            if len(value) > 1:
                val_node = build_list([code(v) for v in value])
            elif len(value) > 0:
                val_node = code(value[0])
            else:
                val_node = para('')
        else:
            val_node = code(value)
        return val_node

    def format_single_value_file(value, package):
        val_list = [code(value)]
        if package:
            val_list.append(nodes.inline(text=', '))
            val_list.append(code(package))
        return nodes.paragraph('', '', *val_list)

    def format_value_file(is_list, value, package):
        if is_list:
            if len(value) > 1:
                val_node = build_list([
                    format_single_value_file(v, p) for v, p in zip(value, package)])
            elif len(value) > 0:
                return format_single_value_file(value[0], package[0])
            else:
                val_node = para('')
        else:
            val_node = format_single_value_file(value, package)
        return val_node

    schema = Schema(rooted_cfg)
    for kp in schema.allkeys():
        if skip_zero_weight and \
           len(kp) == 6 and kp[0] == 'flowgraph' and kp[-2] == 'weight' and \
           schema.get(*kp) == 0:
            continue

        values = schema._getvals(*kp, return_defvalue=False)
        if values:
            # take first of multiple possible values
            value, step, index = values[0]
            val_type = schema.get(*kp, field='type')
            is_filedir = 'file' in val_type or 'dir' in val_type
            # Don't display false booleans
            if val_type == 'bool' and value is False:
                continue
            if is_filedir:
                val_node = format_value_file(val_type.startswith('['), value,
                                             schema.get(*kp, field='package',
                                                        step=step, index=index))
            else:
                val_node = format_value(val_type.startswith('['), value)

            # HTML builder fails if we don't make a text node the parent of the
            # reference node returned by keypath()
            p = nodes.paragraph()
            p += keypath(kp, refdoc)
            table.append([p, val_node])

    if len(table) > 1:
        # This colspec creates two columns of equal width that fill the entire
        # page, and adds line breaks if table cell contents are longer than one
        # line. "\X" is defined by Sphinx, otherwise this is standard LaTeX.
        colspec = r'{|\X{1}{2}|\X{1}{2}|}'
        return build_table(table, colspec=colspec)
    else:
        return None


def build_package_table(schema):
    def collect_packages(cfg):
        packages = []
        if Schema._is_leaf(cfg):
            if 'dir' in cfg['type'] or 'file' in cfg['type']:
                for _, index_data in cfg['node'].items():
                    for _, data in index_data.items():
                        packages.extend(data['package'])
        else:
            for key in cfg:
                packages.extend(collect_packages(cfg[key]))
        packages = [p for p in packages if p]
        return list(set(packages))

    schema = Schema(cfg=schema)
    packages = collect_packages(schema.cfg)

    if not packages:
        return None

    # This colspec creates two columns of equal width that fill the entire
    # page, and adds line breaks if table cell contents are longer than one
    # line. "\X" is defined by Sphinx, otherwise this is standard LaTeX.
    colspec = r'{|\X{1}{2}|\X{1}{2}|}'

    table = [[strong('Package'), strong('Specifications')]]
    for package in packages:
        path = schema.get('package', 'source', package, 'path')
        ref = schema.get('package', 'source', package, 'ref')

        specs = [nodes.paragraph('', 'Path: ', code(path))]
        if ref:
            specs.append(nodes.paragraph('', 'Reference: ', code(ref)))

        table.append([para(package), build_list(specs)])

    return build_table(table, colspec=colspec)


#############
# Base class
#############


def flag_opt(argument):
    if argument is not None:
        raise ValueError('Flag should not have content')
    return True


class DynamicGen(SphinxDirective):
    '''Base class for all three directives provided by this extension.

    Each child class implements a directive by overriding the display_config()
    method and setting a PATH member variable.
    '''

    option_spec = {'nobuiltins': flag_opt}

    def document_module(self, module, modname, path):
        '''Build section documenting given module and name.'''
        print(f'Generating docs for module {modname}...')

        s = build_section_with_target(modname, self.get_document_ref_key(modname),
                                      self.state.document)

        # Attempt to use module doc string first
        if not self.generate_documentation_from_object(module, path, s):
            setup = self.get_setup_method(module)
            # Then use setup doc string
            self.generate_documentation_from_object(setup, path, s)

        chips = None
        try:
            chips = self.configure_chip_for_docs(module)
        except Exception as e:
            print("Failed:", e)
            # raise e

        if not chips:
            return None

        if not isinstance(chips, list):
            chips = [chips]

        for chip in chips:
            extra_content = self.extra_content(chip, modname)
            if extra_content is not None:
                s += extra_content

            package_info = self.package_information(chip, modname)
            if package_info is not None:
                s += [package_info]

            disp = self.display_config(chip, modname)
            if disp:
                s += disp

        child_content = self.child_content(path, module, modname)
        if child_content is not None:
            s += child_content

        return s

    def run(self):
        '''Main entry point of directive.'''
        sections = []
        self.env.note_dependency(__file__)
        self.env.note_dependency(utils.__file__)

        for module, modname in self.get_modules():
            path = module.__file__
            self.env.note_dependency(path)
            docs = self.document_module(module, modname, path)
            if docs is not None:
                sections.append((docs, modname))

        if len(sections) > 0:
            # Sort sections by module name
            sections = sorted(sections, key=lambda t: t[1])
            # Strip off modname so we just return list of docutils sections
            sections, _ = zip(*sections)

        return list(sections)

    def get_modules(self):
        '''Gets dynamic modules under `self.PATH`.

        This function explicitly searches builtins.
        '''

        modules = []
        for plugin in get_plugins("docs", name=self.PATH):
            for mod in plugin():
                if isinstance(mod, str):
                    modules.extend(self.get_modules_in_dir(mod))
                else:
                    modules.append(mod)

        return modules

    def get_modules_in_dir(self, module_dir):
        '''Routine for getting modules and their names from a certain
        directory.'''
        modules = []
        for importer, modname, _ in pkgutil.iter_modules([module_dir]):
            module = importer.find_spec(modname).loader.load_module(modname)
            modules.append((module, modname))

        return modules

    def parse_rst(self, content, s):
        '''Helper for parsing reStructuredText content, adding it directly to
        section `s`.'''
        rst = ViewList()
        # use fake filename 'inline' for error # reporting
        for i, line in enumerate(content.split('\n')):
            rst.append(line, 'inline', i)
        nested_parse_with_titles(self.state, rst, s)

    def package_information(self, chip, modname):
        packages = build_package_table(chip.schema.cfg)
        if packages:
            sec = build_section('Data sources', self.get_data_source_ref_key(modname, chip.design))
            sec += packages
            return sec

        return None

    def extra_content(self, chip, modname):
        '''Adds extra content to documentation.

        May return a list of docutils nodes that will be added to the
        documentation in between a module's docstring and configuration table.
        Otherwise, if return value is None, don't add anything.
        '''
        return None

    def child_content(self, path, module, modname):
        return None

    def generate_documentation_from_object(self, func, path, s):
        # raw docstrings have funky indentation (basically, each line is already
        # indented as much as the function), so we call trim() helper function
        # to clean it up
        docstr = utils.trim(func.__doc__)

        if docstr:
            self.parse_rst(docstr, s)
        else:
            return False

        src_link = None
        for docs_link in get_plugins("docs", name="linkcode"):
            src_link = docs_link(file=path)
            if src_link:
                break

        if src_link:
            p = para('Setup file: ')
            p += link(src_link, text=os.path.basename(path))
            s += p

        return True

    def _document_free_params(self, cfg, type, key_path, reference_prefix, s):
        if type in cfg:
            cfg = cfg[type]
        else:
            return

        if type == "var":
            type_heading = "Variables"
        elif type == "file":
            type_heading = "Files"
        elif type == "dir":
            type_heading = "Directories"

        table = [[strong('Parameters'), strong('Help')]]
        for key, params in cfg.items():
            if key == "default":
                continue

            key_node = nodes.paragraph()
            key_node += keypath(key_path + [key], self.env.docname,
                                key_text=["...", f"'{type}'", f"'{key}'"])
            table.append([key_node, para(params["help"])])

        if len(table) > 1:
            s += build_section(type_heading, self.get_ref(*reference_prefix, type))
            colspec = r'{|\X{1}{2}|\X{1}{2}|}'
            s += build_table(table, colspec=colspec)

    def get_make_docs_method(self, module):
        return getattr(module, 'make_docs', None)

    def get_configure_docs_method(self, module):
        return getattr(module, 'configure_docs', None)

    def get_setup_method(self, module):
        return getattr(module, 'setup', None)

    def make_chip(self):
        return siliconcompiler.Chip('<design>')

    def _handle_make_docs(self, chip, module):
        make_docs = self.get_make_docs_method(module)
        if make_docs:
            new_chip = make_docs(chip)
            if new_chip:
                # make_docs returned something so it's fully configured
                return (new_chip, True)
            else:
                return (chip, False)
        return (None, False)

    def _handle_setup(self, chip, module):
        setup = self.get_setup_method(module)
        if not setup:
            return None
        new_chip = setup()
        if new_chip:
            return new_chip
        else:
            # Setup didn't return anything so return the Chip object
            return chip

    def configure_chip_for_docs(self, module):
        chip = self.make_chip()
        docs_chip, docs_configured = self._handle_make_docs(chip, module)
        if docs_chip and docs_configured:
            return docs_chip

        return self._handle_setup(chip, module)

    def get_ref_prefix(self):
        return self.REF_PREFIX

    @staticmethod
    def get_ref_key(*path):
        return '-'.join(path)

    def get_ref(self, *sections):
        return DynamicGen.get_ref_key(self.get_ref_prefix(), *sections)

    def get_configuration_ref_key(self, *modname):
        return self.get_ref(*modname, 'configuration')

    def get_data_source_ref_key(self, *modname):
        return self.get_ref(*modname, 'data_source')

    def get_document_ref_key(self, *modname):
        return self.get_ref(*modname)

    def build_config_recursive(self, schema, refdoc, keypath=None, sec_key_prefix=None):
        '''Helper function for displaying schema at each level as tables under nested
        sections.

        For each item:
        - If it's a leaf, collect it into a table we will display at this
            level
        - Otherwise, recurse and collect sections of lower levels
        '''
        if keypath is None:
            keypath = []
        if sec_key_prefix is None:
            sec_key_prefix = []

        leaves = {}
        child_sections = []
        for key in schema.getkeys(*keypath):
            if Schema._is_leaf(schema.getdict(*keypath, key)):
                val = schema.getdict(*keypath, key)
                leaves.update({key: val})
            else:
                children = self.build_config_recursive(
                    schema,
                    refdoc,
                    keypath=keypath + [key],
                    sec_key_prefix=sec_key_prefix)
                child_sections.extend(children)

        schema_table = None
        if len(leaves) > 0:
            # Might return None is none of the leaves are displayable
            schema_table = build_schema_value_table(leaves, refdoc, keypath_prefix=keypath)

        if schema_table is not None:
            # If we've found leaves, create a new section where we'll display a
            # table plus all child sections.
            keypathstr = ', '.join(keypath)
            top = build_section(keypathstr, self.get_ref(*sec_key_prefix, 'key', *keypath))
            top += schema_table
            top += child_sections
            return [top]

        # Otherwise, just pass on the child sections -- we don't want to
        # create an extra level of section hierarchy for levels of the
        # schema without leaves.
        return child_sections

#########################
# Specialized extensions
#########################


class FlowGen(DynamicGen):
    PATH = 'flows'
    REF_PREFIX = 'flows'

    def extra_content(self, chip, modname):
        flow_path = os.path.join(self.env.app.outdir, f'_images/gen/{modname}.svg')
        chip.write_flowgraph(flow_path, flow=chip.design)
        return [image(flow_path, center=True)]

    def display_config(self, chip, modname):
        '''Display parameters under `flowgraph, <step>`, `metric, <step>` and
        `showtool`. Parameters are grouped into sections by step, with an
        additional table for non-step items.
        '''

        name = chip.design

        settings = build_section('Configuration', self.get_configuration_ref_key(name))

        steps = chip.getkeys('flowgraph', chip.design)
        # TODO: should try to order?

        # Build section + table for each step (combining entries under flowgraph
        # and metric)
        for step in steps:
            section = build_section(step, self.get_ref(name, 'step', step))
            step_cfg = {}
            cfg = chip.getdict('flowgraph', chip.design, step)
            if cfg is None:
                continue
            schema = Schema(cfg=cfg)
            schema.prune()
            pruned = schema.cfg
            if chip.design not in step_cfg:
                step_cfg[chip.design] = {}
            step_cfg[chip.design][step] = pruned

            section += build_schema_value_table(step_cfg,
                                                self.env.docname,
                                                keypath_prefix=['flowgraph'],
                                                skip_zero_weight=True)
            settings += section

        return settings


class PDKGen(DynamicGen):
    PATH = 'pdks'
    REF_PREFIX = 'pdks'

    def display_config(self, chip, modname):
        '''Display parameters under `pdk`, `asic`, and `library` in nested form.'''

        name = chip.design

        settings = build_section('Configuration', self.get_configuration_ref_key(name))

        config_settings = self.build_config_recursive(
            chip.schema,
            self.env.docname,
            keypath=['pdk'],
            sec_key_prefix=[name])

        if config_settings:
            settings += config_settings
            return settings

        return None


class LibGen(DynamicGen):
    PATH = 'libs'
    REF_PREFIX = 'libs'

    def extra_content(self, chip, modname):
        # assume same pdk for all libraries configured by this module
        pdk = chip.get('option', 'pdk')

        p = docutils.nodes.inline('')
        pdk_ref = "None"
        if pdk:
            pdkid = get_ref_id(DynamicGen.get_ref_key(PDKGen.REF_PREFIX, pdk))
            pdk_ref = f":ref:`{pdk}<{pdkid}>`"
        self.parse_rst(f'Associated PDK: {pdk_ref}', p)

        return [p]

    def display_config(self, chip, modname):
        '''Display parameters under in nested form.'''

        sections = []

        libname = chip.design

        settings = build_section_with_target(libname, self.get_ref(libname),
                                             self.state.document)

        for key in ('asic', 'input', 'output', 'option'):
            settings += self.build_config_recursive(
                chip.schema,
                self.env.docname,
                keypath=[key],
                sec_key_prefix=[libname, key])

        sections.append(settings)

        return sections

    def get_document_ref_key(self, *modname):
        return super().get_document_ref_key(*modname, "grouping")


class ToolGen(DynamicGen):
    PATH = 'tools'
    REF_PREFIX = 'tools'

    def make_chip(self):
        chip = super().make_chip()
        self._setup_chip(chip, '<tool>', '<task>')

        return chip

    def _setup_chip(self, chip, tool_name, task_name):
        step = chip.get('arg', 'step')
        if not step:
            step = '<step>'
        index = chip.get('arg', 'index')
        if not index:
            index = '<index>'
        chip.set('arg', 'step', step)
        chip.set('arg', 'index', index)

        flow = chip.get('option', 'flow')
        if not flow:
            flow = '<flow>'
        chip.set('option', 'flow', flow)
        chip.set('flowgraph', flow, step, index, 'tool', tool_name)
        chip.set('flowgraph', flow, step, index, 'task', task_name)

    def configure_chip_for_docs(self, module, toolmodule=None):
        chip = self.make_chip()
        docs_chip, docs_configured = self._handle_make_docs(chip, module)
        if not docs_chip and toolmodule:
            docs_chip, docs_configured = self._handle_make_docs(chip, toolmodule)
        if docs_configured:
            return docs_chip

        # set values for current step
        toolname = module.__name__
        if self.__tool:
            toolname = self.__tool
        taskname = '<task>'
        if self.__task:
            taskname = self.__task
        self._setup_chip(chip, toolname, taskname)

        if toolmodule:
            return chip
        else:
            return self._handle_setup(chip, module)

    def display_config(self, chip, modname):
        '''Display config under `eda, <modname>` in a single table.'''
        cfg = chip.getdict('tool', modname)
        schema = Schema(cfg=cfg)
        schema.prune()
        pruned = schema.cfg
        if 'task' in pruned:
            # Remove task specific items since they will be documented
            # by the task documentation
            del pruned['task']
        table = build_schema_value_table(pruned, self.env.docname, keypath_prefix=['tool', modname])
        if table is not None:
            return table
        else:
            return []

    def task_display_config(self, chip, toolname, taskname):
        '''Display config under `eda, <modname>` in a single table.'''
        cfg = chip.getdict('tool', toolname, 'task', taskname)
        schema = Schema(cfg=cfg)
        for vals, step, index in schema._getvals('require'):
            schema.set('require', sorted(set(vals)),
                       step=step, index=index)
        schema.prune()
        pruned = schema.cfg
        table = build_schema_value_table(pruned, self.env.docname,
                                         keypath_prefix=['tool', toolname, 'task', taskname])
        if table is not None:
            return table
        else:
            return []

    def get_modules_in_dir(self, module_dir):
        self.__tool = None
        self.__task = None
        '''Custom implementation for ToolGen since the tool setup modules are
        under an extra directory, and this way we don't have to force users to
        add an __init__.py to make the directory a module itself.
        '''
        modules = []
        for toolname in os.listdir(module_dir):
            if (toolname == "template" or toolname.startswith('_')):
                # No need to include empty template in documentation
                continue
            # skip over directories/files that don't match the structure of tool
            # directories (otherwise we'll get confused by Python metadata like
            # __pycache__/)
            if not os.path.isdir(f'{module_dir}/{toolname}'):
                continue
            path = f'{module_dir}/{toolname}/{toolname}.py'
            if not os.path.exists(path):
                path = f'{module_dir}/{toolname}/__init__.py'
            if not os.path.exists(path):
                continue

            spec = importlib.util.spec_from_file_location(toolname, path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            modules.append((module, toolname))

        return modules

    def document_task(self, path, toolmodule, taskmodule, taskname, toolname):
        self.env.note_dependency(path)
        s = build_section_with_target(taskname, self.get_ref(toolname, taskname),
                                      self.state.document)

        # Find setup function
        task_setup = self.get_setup_method(taskmodule)
        if not task_setup:
            return None

        print(f"Generating docs for task {toolname}/{taskname}...")

        self.generate_documentation_from_object(task_setup, path, s)

        self.__tool = toolname
        self.__task = taskname
        chip = self.configure_chip_for_docs(taskmodule, toolmodule=toolmodule)
        self.__tool = None
        self.__task = None

        # Annotate the target used for default values
        if chip.valid('option', 'target') and chip.get('option', 'target'):
            p = docutils.nodes.inline('')
            target = chip.get('option', 'target').split('.')[-1]
            targetid = get_ref_id(DynamicGen.get_ref_key(TargetGen.REF_PREFIX, target))
            self.parse_rst(f"Built using target: :ref:`{target}<{targetid}>`", p)
            s += p

        try:
            task_setup(chip)

            config = build_section("Configuration", self.get_configuration_ref_key(toolname,
                                                                                   taskname))
            config_table = self.task_display_config(chip, toolname, taskname)
            if config_table:
                s += config
                s += config_table
            self.document_free_params(chip.getdict('tool', toolname, 'task', taskname),
                                      [toolname, taskname, 'params'],
                                      s)
        except Exception as e:
            print(f'Failed to document task, Chip object probably not configured correctly: {e}')
            return None

        return s

    def child_content(self, path, module, modname):
        sections = []
        path = os.path.abspath(path)
        module_dir = os.path.dirname(path)
        for taskfile in os.listdir(module_dir):
            if taskfile.startswith("_"):
                # skip private file
                continue

            task_path = os.path.join(module_dir, taskfile)
            if path == task_path:
                # skip tool module
                continue

            if not os.path.isfile(task_path):
                # skip if not a file
                continue

            spec = importlib.util.spec_from_file_location(taskfile, task_path)
            if not spec:
                # unable to load, probably not a python file
                continue
            taskmodule = importlib.util.module_from_spec(spec)
            try:
                spec.loader.exec_module(taskmodule)
            except Exception:
                # Module failed to load
                # klayout imports pya which is only defined in klayout
                continue

            taskname = os.path.splitext(os.path.basename(task_path))[0]

            task_doc = self.document_task(task_path, module, taskmodule, taskname, modname)
            if task_doc:
                sections.append((taskname, task_doc))

        if len(sections) > 0:
            sections = sorted(sections, key=lambda t: t[0])
            # Strip off modname so we just return list of docutils sections
            _, sections = zip(*sections)
        return sections

    def document_free_params(self, cfg, reference_prefix, s):
        key_path = ['tool', '<tool>', 'task', '<task>']
        self._document_free_params(cfg, 'var', key_path + ['var'], reference_prefix, s)
        self._document_free_params(cfg, 'file', key_path + ['file'], reference_prefix, s)
        self._document_free_params(cfg, 'dir', key_path + ['dir'], reference_prefix, s)

    def _handle_setup(self, chip, module):
        setup = self.get_setup_method(module)
        if not setup:
            return None

        setup(chip)

        return chip


class TargetGen(DynamicGen):
    PATH = 'targets'
    REF_PREFIX = 'targets'

    def build_module_list(self, chip, header, modtype, targetname, *refprefix):
        modules = chip._loaded_modules[modtype]
        if len(modules) > 0:
            section = build_section(header, self.get_ref(targetname, modtype))
            modlist = nodes.bullet_list()
            for module in modules:
                list_item = nodes.list_item()
                # TODO: replace with proper docutils nodes: sphinx.addnodes.pending_xref
                modkey = get_ref_id(DynamicGen.get_ref_key(*refprefix, module))
                self.parse_rst(f':ref:`{module}<{modkey}>`', list_item)
                modlist += list_item

            section += modlist
            return section
        return None

    def display_config(self, chip, modname):
        sections = []

        pdk_section = self.build_module_list(
            chip, 'Included PDK', 'pdks', modname, PDKGen.REF_PREFIX)
        if pdk_section is not None:
            sections.append(pdk_section)

        libs_section = self.build_module_list(
            chip, 'Included libraries', 'libs', modname, LibGen.REF_PREFIX)
        if libs_section is not None:
            sections.append(libs_section)

        flow_section = self.build_module_list(
            chip, 'Included flows', 'flows', modname, FlowGen.REF_PREFIX)
        if flow_section is not None:
            sections.append(flow_section)

        checklist_section = self.build_module_list(
            chip, 'Included checklists', 'checklists', modname, ChecklistGen.REF_PREFIX)
        if checklist_section is not None:
            sections.append(checklist_section)

        filtered_cfg = {}
        for key in ('asic', 'constraint', 'option'):
            filtered_cfg[key] = chip.getdict(key)
        schema = Schema(cfg=filtered_cfg)
        schema.prune()
        pruned_cfg = schema.cfg

        if len(pruned_cfg) > 0:
            schema_section = build_section('Configuration', self.get_configuration_ref_key(modname))
            schema_section += build_schema_value_table(pruned_cfg, self.env.docname)
            sections.append(schema_section)

        return sections

    def _handle_setup(self, chip, module):
        setup = self.get_setup_method(module)
        if not setup:
            return None

        setup(chip)

        return chip


class AppGen(DynamicGen):
    PATH = 'apps'
    REF_PREFIX = 'apps'

    def document_module(self, module, modname, path):
        if modname[0] == "_":
            return None

        cmd_name = modname.replace('_', '-')
        cmd = [cmd_name, '--help']

        output = subprocess.check_output(cmd).decode('utf-8')

        section = build_section(cmd_name, self.get_ref(cmd_name))
        section += literalblock(output)

        return section


class ChecklistGen(DynamicGen):
    PATH = 'checklists'
    REF_PREFIX = 'checklists'

    def display_config(self, chip, modname):
        '''Display parameters under in nested form.'''

        sections = []

        name = chip.design

        keypath_prefix = ['checklist', name]
        cfg = chip.getdict(*keypath_prefix)

        settings = build_section('Configuration', self.get_configuration_ref_key(name))

        for key in cfg.keys():
            if key == 'default':
                continue
            settings += build_section(key, self.get_ref(name, 'key', key))
            settings += build_schema_value_table(cfg[key], self.env.docname,
                                                 keypath_prefix=[*keypath_prefix, key])

        sections.append(settings)

        return sections


class ExampleGen(DynamicGen):

    def get_modules(self):
        examples_dir = f'{SC_ROOT}/examples'

        modules = []
        for example in os.listdir(examples_dir):
            if not os.path.isdir(f'{examples_dir}/{example}'):
                continue
            path = f'{examples_dir}/{example}/{example}.py'
            if not os.path.exists(path):
                continue

            spec = importlib.util.spec_from_file_location(example, path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            modules.append((module, example))

        return modules

    def document_module(self, module, modname, path):
        section = build_section(modname, modname)

        if not hasattr(module, 'main'):
            return None

        main = getattr(module, 'main')

        # raw docstrings have funky indentation (basically, each line is already
        # indented as much as the function), so we call trim() helper function
        # to clean it up
        docstr = utils.trim(main.__doc__)

        if docstr:
            self.parse_rst(docstr, section)

        return section


def keypath_role(name, rawtext, text, lineno, inliner, options=None, content=None):
    doc = inliner.document
    env = doc.settings.env

    # Split and clean up keypath
    keys = [key.strip() for key in text.split(',')]
    try:
        return [keypath(keys, env.docname)], []
    except ValueError as e:
        msg = inliner.reporter.error(f'{rawtext}: {e}', line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]


class SCDomain(StandardDomain):
    name = 'sc'

    # Override in StandardDomain so xref is literal instead of inline
    # https://github.com/sphinx-doc/sphinx/blob/ba080286b06cb9e0cadec59a6cf1f96aa11aef5a/sphinx/domains/std.py#L789
    def build_reference_node(self,
                             fromdocname,
                             builder,
                             docname,
                             labelid,
                             sectname,
                             rolename,
                             **options):
        nodeclass = options.pop('nodeclass', nodes.reference)
        newnode = nodeclass('', '', internal=True, **options)
        innernode = nodes.literal(sectname, sectname)
        if innernode.get('classes') is not None:
            innernode['classes'].append('std')
            innernode['classes'].append('std-' + rolename)
        if docname == fromdocname:
            newnode['refid'] = labelid
        else:
            # set more info in contnode; in case the
            # get_relative_uri call raises NoUri,
            # the builder will then have to resolve these
            contnode = pending_xref('')
            contnode['refdocname'] = docname
            contnode['refsectname'] = sectname
            newnode['refuri'] = builder.get_relative_uri(
                fromdocname, docname)
            if labelid:
                newnode['refuri'] += '#' + labelid
        newnode.append(innernode)
        return newnode


def setup(app):
    app.add_domain(SCDomain)
    app.add_directive('flowgen', FlowGen)
    app.add_directive('pdkgen', PDKGen)
    app.add_directive('libgen', LibGen)
    app.add_directive('toolgen', ToolGen)
    app.add_directive('appgen', AppGen)
    app.add_directive('examplegen', ExampleGen)
    app.add_directive('targetgen', TargetGen)
    app.add_directive('checklistgen', ChecklistGen)
    app.add_role('keypath', keypath_role)

    return {
        'version': siliconcompiler.__version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



import os

from siliconcompiler.tools.gtkwave import setup as tool_setup
from siliconcompiler.tools._common import \
    add_require_input, get_tool_task, input_provides
from siliconcompiler import utils


def setup(chip):
    '''
    Show a VCD file.
    '''

    tool_setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'refdir',
             'tools/gtkwave/scripts',
             step=step, index=index, package='siliconcompiler')

    chip.set('tool', tool, 'task', task, 'refdir',
             'tools/gtkwave/scripts',
             step=step, index=index, package='siliconcompiler')
    chip.set('tool', tool, 'task', task, 'script', 'sc_show.tcl',
             step=step, index=index)

    if f'{chip.top()}.vcd' in input_provides(chip, step, index):
        chip.set('tool', tool, 'task', task, 'input', f'{chip.top()}.vcd', step=step, index=index)
    elif chip.valid('tool', tool, 'task', task, 'var', 'show_filepath') and \
            chip.get('tool', tool, 'task', task, 'var', 'show_filepath', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'show_filepath']),
                 step=step, index=index)
    else:
        add_require_input(chip, 'input', 'waveform', 'vcd')

    chip.set('tool', tool, 'task', task, 'var', 'show_exit', False,
             step=step, index=index, clobber=False)


def runtime_options(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    options = []

    threads = chip.get('tool', tool, 'task', task, 'threads', step=step, index=index)
    if threads:
        options.append(f'--cpu={threads}')

    script = chip.find_files('tool', tool, 'task', task, 'script', step=step, index=index)[0]
    options.append(f'--script={script}')

    if os.path.exists(f'inputs/{chip.top()}.vcd'):
        dump = f'inputs/{chip.top()}.vcd'
    elif chip.valid('tool', tool, 'task', task, 'var', 'show_filepath') and \
            chip.get('tool', tool, 'task', task, 'var', 'show_filepath', step=step, index=index):
        dump = chip.get('tool', tool, 'task', task, 'var', 'show_filepath',
                        step=step, index=index)[0]
    else:
        dump = chip.find_files('input', 'waveform', 'vcd', step=step, index=index)[0]
    options.append(f'--dump={dump}')

    return options



# Copyright 2020 Silicon Compiler Authors. All Rights Reserved.

import sys

from siliconcompiler.remote.server import Server


###############################################
# Main method to run the sc-server application.
###############################################
def main():
    progname = "sc-server"
    description = """
-----------------------------------------------------------
Silicon Compiler Collection Remote Job Server (sc-server)
-----------------------------------------------------------
"""

    server = Server()

    try:
        server.create_cmdline(
            progname,
            description=description)
    except ValueError as e:
        server.logger.error(f'{e}')
        return 1

    try:
        server.run()
    except Exception as e:
        server.logger.error(f'{e}')
        return 1

    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2020 Silicon Compiler Authors. All Rights Reserved.
import sys
import os
import siliconcompiler
from siliconcompiler.utils import get_default_iomap
from siliconcompiler.apps._common import manifest_switches, pick_manifest, UNSET_DESIGN
from siliconcompiler.utils import get_file_ext


def main():
    progname = "sc-show"
    description = """
    --------------------------------------------------------------
    Restricted SC app that displays the layout of a design
    based on a file provided or tries to display the final
    layout based on loading the json manifest from:
    build/<design>/job0/<design>.pkg.json

    Examples:

    sc-show
    (displays build/adder/job0/write.gds/0/outputs/adder.gds)

    sc-show -design adder
    (displays build/adder/job0/write.gds/0/outputs/adder.gds)

    sc-show -design adder -arg_step floorplan
    (displays build/adder/job0/floorplan/0/outputs/adder.def)

    sc-show -design adder -arg_step place -arg_index 1
    (displays build/adder/job0/place/1/outputs/adder.def)

    sc-show -design adder -jobname rtl2gds
    (displays build/adder/rtl2gds/write.gds/0/outputs/adder.gds)

    sc-show -cfg build/adder/rtl2gds/adder.pkg.json
    (displays build/adder/rtl2gds/write.gds/0/outputs/adder.gds)

    sc-show -design adder -ext odb
    (displays build/adder/job0/write.views/0/outputs/adder.odb)

    sc-show build/adder/job0/route/1/outputs/adder.def
    (displays build/adder/job0/route/1/outputs/adder.def)
    """

    # Create a base chip class.
    chip = siliconcompiler.Chip(UNSET_DESIGN)

    # Fill input map with default mapping only for showable files
    input_map = {}
    default_input_map = get_default_iomap()
    for ext in chip._showtools:
        if ext in default_input_map:
            input_map[ext] = default_input_map[ext]

    extension_arg = {
        'metavar': '<ext>',
        'help': '(optional) Specify the extension of the file to show.',
        'sc_print': False
    }
    screenshot_arg = {
        'action': 'store_true',
        'help': '(optional) Will generate a screenshot and exit.',
        'sc_print': False
    }

    try:
        args = chip.create_cmdline(
            progname,
            switchlist=[*manifest_switches(),
                        '-input',
                        '-loglevel'],
            description=description,
            input_map=input_map,
            additional_args={
                '-ext': extension_arg,
                '-screenshot': screenshot_arg
            })
    except Exception as e:
        chip.logger.error(e)
        return 1

    # Search input keys for files
    input_mode = []
    for fileset in chip.getkeys('input'):
        for mode in chip.getkeys('input', fileset):
            if chip.schema._getvals('input', fileset, mode):
                input_mode = [('input', fileset, mode)]

    filename = None
    if input_mode:
        check_ext = list(chip._showtools.keys())

        if args['ext']:
            check_ext = [args['ext']]

        def get_file_from_keys():
            for ext in check_ext:
                for key in input_mode:
                    for files, _, _ in chip.schema._getvals(*key):
                        for file in files:
                            if get_file_ext(file) == ext:
                                return file
            return None

        filename = get_file_from_keys()

    # Attempt to load a manifest
    if not chip.get('option', 'cfg'):
        manifest = pick_manifest(chip, src_file=filename)
        if manifest:
            chip.logger.info(f'Loading manifest: {manifest}')
            chip.read_manifest(manifest)

    # Error checking
    design = chip.get('design')
    design_set = design != UNSET_DESIGN
    if not (design_set or input_mode):
        chip.logger.error('Nothing to load: please define a target with '
                          '-cfg, -design, and/or inputs.')
        return 1

    # Read in file
    if filename:
        chip.logger.info(f"Displaying {filename}")

    if not chip.find_files('option', 'builddir', missing_ok=True):
        chip.logger.warning("Unable to access original build directory "
                            f"\"{chip.get('option', 'builddir')}\", using \"build\" instead")
        chip.set('option', 'builddir', 'build')

    success = chip.show(filename,
                        extension=args['ext'],
                        screenshot=args['screenshot'])

    if args['screenshot'] and os.path.isfile(success):
        chip.logger.info(f'Screenshot file: {success}')

    return 0 if success else 1


#########################
if __name__ == "__main__":
    sys.exit(main())



'''
slang is a software library that provides various components for lexing, parsing, type checking,
and elaborating SystemVerilog code. It comes with an executable tool that can compile and lint
any SystemVerilog project, but it is also intended to be usable as a front end for synthesis
tools, simulators, linters, code editors, and refactoring tools.

Documentation: https://sv-lang.com/

Sources: https://github.com/MikePopoloski/slang

Installation: https://sv-lang.com/building.html
'''
import re
from siliconcompiler import sc_open
from siliconcompiler.tools._common import record_metric
from siliconcompiler.tools._common import \
    get_frontend_options, get_input_files, get_tool_task


################################
# Setup Tool (pre executable)
################################
def setup(chip):
    tool = 'slang'

    # Standard Setup
    chip.set('tool', tool, 'exe', 'slang')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=6.0', clobber=False)


def parse_version(stdout):
    # slang --version output looks like:
    # slang version 6.0.121+c2c478cf

    # grab version # by splitting on whitespace
    return stdout.strip().split()[-1].split('+')[0]


def post_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    log = f'{step}.log'
    with sc_open(log) as f:
        for line in f:
            match = re.search(r'(\d+) errors, (\d+) warnings', line)
            if match:
                record_metric(chip, step, index, 'errors', match.group(1), log)
                record_metric(chip, step, index, 'warnings', match.group(2), log)


def common_runtime_options(chip):
    options = []

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    options.extend(['-j', str(chip.get('tool', tool, 'task', task, 'threads',
                              step=step, index=index))])

    opts = get_frontend_options(chip,
                                ['ydir',
                                 'idir',
                                 'vlib',
                                 'libext',
                                 'define',
                                 'param'])

    if opts['libext']:
        options.append(f'--libext {",".join(opts["libext"])}')

    #####################
    # Library directories
    #####################
    if opts['ydir']:
        options.append(f'-y {",".join(opts["ydir"])}')

    #####################
    # Library files
    #####################
    if opts['vlib']:
        options.append(f'-libfile {",".join(opts["vlib"])}')

    #####################
    # Include paths
    #####################
    if opts['idir']:
        options.append(f'--include-directory {",".join(opts["idir"])}')

    #######################
    # Variable Definitions
    #######################
    for value in opts['define']:
        options.append('-D ' + value)

    #######################
    # Command files
    #######################
    cmdfiles = get_input_files(chip, 'input', 'cmdfile', 'f')
    if cmdfiles:
        options.append(f'-F {",".join(cmdfiles)}')

    #######################
    # Sources
    #######################
    for value in get_input_files(chip, 'input', 'rtl', 'systemverilog'):
        options.append(value)
    for value in get_input_files(chip, 'input', 'rtl', 'verilog'):
        options.append(value)

    #######################
    # Top Module
    #######################
    options.append('--top ' + chip.top())

    ###############################
    # Parameters (top module only)
    ###############################
    # Set up user-provided parameters to ensure we elaborate the correct modules
    for param, value in opts['param']:
        value = value.replace('"', '\\"')
        options.append(f'-G {param}={value}')

    return options
